# coding=utf-8
"""
Implementation for:
- UEVaultManagerCLI: command line interface for UEVaultManager
"""
import argparse
import csv
import json
import logging
import os
import shutil
import subprocess
import sys
import time
import webbrowser
from collections import namedtuple
from datetime import datetime
from logging.handlers import QueueListener
from multiprocessing import freeze_support, Queue as MPQueue
from platform import platform
import UEVaultManager.tkgui.modules.functions_no_deps as gui_fn  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
# noinspection PyPep8Naming
from UEVaultManager import __version__ as UEVM_version, __codename__ as UEVM_codename
from UEVaultManager.api.egs import create_empty_assets_extras, GrabResult
from UEVaultManager.api.uevm import UpdateSeverity
from UEVaultManager.core import AppCore
from UEVaultManager.models.csv import CSV_headings
from UEVaultManager.models.exceptions import InvalidCredentialsError
from UEVaultManager.tkgui.main import init_gui
from UEVaultManager.tkgui.modules.DisplayContentWindowClass import DisplayContentWindow
from UEVaultManager.tkgui.modules.functions import custom_print  # simplier way to use the custom_print function
from UEVaultManager.tkgui.modules.functions import json_print_key_val
from UEVaultManager.tkgui.modules.ProgressWindowsClass import ProgressWindow
from UEVaultManager.tkgui.modules.SaferDictClass import SaferDict
from UEVaultManager.tkgui.modules.UEVMGuiClass import UEVMGui
from UEVaultManager.tkgui.modules.UEVMGuiHiddenRootClass import UEVMGuiHiddenRoot
from UEVaultManager.utils.cli import str_to_bool, check_and_create_path, create_list_from_string, str_is_bool
from UEVaultManager.utils.custom_parser import HiddenAliasSubparsersAction

logging.basicConfig(format='[%(name)s] %(levelname)s: %(message)s', level=logging.INFO)


def init_gui_args(args, additional_args=None) -> None:
    """
    Initialize the GUI arguments using the CLI arguments
    :param args:
    :param additional_args: dict of additional arguments to add
    """

    # args can not be used as it because it's an object that mainly run as a dict (but it's not)
    # so we need to convert it to a dict first
    temp_dict = vars(args)
    temp_dict['csv'] = True  # force csv output
    temp_dict['gui'] = True
    if additional_args is not None:
        temp_dict.update(additional_args)
    # create a SaferDict object from the dict (it will avoid errors when trying to access non-existing keys)
    gui_g.UEVM_cli_args = SaferDict({})
    # copy the dict content to the SaferDict object
    gui_g.UEVM_cli_args.copy_from(temp_dict)


def init_progress_window(args, logger=None, callback=None) -> (bool, ProgressWindow):
    """
    Initialize the progress window
    :param args: args of the command line
    :param logger: logger to use
    :param callback: callback function to call while progress updating
    :return: (True if the UEVMGui window already existed | False, ProgressWindow)
    """
    gui_g.UEVM_log_ref = logger

    # check if the GUI is already running
    if gui_g.UEVM_gui_ref is None:
        # create a fake root because ProgressWindow must always be a top level window
        gui_g.UEVM_gui_ref = UEVMGuiHiddenRoot()
        uewm_gui_exists = False
    else:
        uewm_gui_exists = True
    force_refresh = True if args.force_refresh else False
    window = ProgressWindow(
        title="Updating Assets List",
        quit_on_close=not uewm_gui_exists,
        width=300,
        height=150,
        max_value=200,
        show_start_button=False,
        show_stop_button=True,
        function=callback,
        function_parameters={
            'filter_category': gui_g.UEVM_filter_category,
            'force_refresh': force_refresh,
        }
    )
    return uewm_gui_exists, window


def init_display_window(logger=None) -> (bool, DisplayContentWindow):
    """
    Initialize the display window
    :param logger: logger to use
    :return: (True if the UEVMGui window already existed | False, DisplayContentWindow)
    """
    gui_g.UEVM_log_ref = logger

    # check if the GUI is already running
    if gui_g.UEVM_gui_ref is None:
        # create a fake root because DisplayContentWindow must always be a top level window
        gui_g.UEVM_gui_ref = UEVMGuiHiddenRoot()
        uewm_gui_exists = False
    else:
        uewm_gui_exists = True
    if gui_g.display_content_window_ref is None:
        gui_g.display_content_window_ref = DisplayContentWindow(title='UEVM: status command output', quit_on_close=not uewm_gui_exists)
    return uewm_gui_exists, gui_g.display_content_window_ref


class UEVaultManagerCLI:
    """
    Command line interface for UEVaultManager
    :param override_config: path to a config file to use instead of the default one
    :param api_timeout: timeout for API requests
    """

    def __init__(self, override_config=None, api_timeout=None):
        self.core = AppCore(override_config, timeout=api_timeout)
        self.logger = logging.getLogger('Cli')
        self.logging_queue = None

    def setup_threaded_logging(self) -> QueueListener:
        """
        Setup logging for the CLI
        """
        self.logging_queue = MPQueue(-1)
        handler = logging.StreamHandler()
        formatter = logging.Formatter('[%(name)s] %(levelname)s: %(message)s')
        handler.setFormatter(formatter)
        ql = QueueListener(self.logging_queue, handler)
        ql.start()
        return ql

    def _resolve_aliases(self, name: str) -> str:
        """
        Resolve an alias
        :param name: name to resolve
        :return: real name
        """
        # make sure aliases exist if not yet created
        self.core.update_aliases(force=False)
        name = name.strip()
        # resolve alias (if any) to real app name
        return self.core.uevmlfs.config.get(section='UEVaultManager.aliases', option=name, fallback=self.core.uevmlfs.aliases.get(name.lower(), name))

    @staticmethod
    def _print_json(data, pretty=False):
        if pretty:
            print(json.dumps(data, indent=2, sort_keys=True))
        else:
            print(json.dumps(data))

    def create_file_backup(self, file_src: str) -> None:
        """
        Create a backup of a file
        :param file_src: path to the file to back up
        """
        # make a backup of the existing file

        # for files defined relatively to the config folder
        file_src = file_src.replace('~/.config', self.core.uevmlfs.path)

        if not file_src:
            return
        try:
            file_name_no_ext, file_ext = os.path.splitext(file_src)
            file_backup = f'{file_name_no_ext}.BACKUP_{datetime.now().strftime("%y-%m-%d_%H-%M-%S")}{file_ext}'
            shutil.copy(file_src, file_backup)
            self.logger.info(f'File {file_src} has been copied to {file_backup}')
        except FileNotFoundError:
            self.logger.info(f'File {file_src} has not been found')

    def create_log_file_backup(self) -> None:
        """
        Create a backup of the log files
        """
        self.create_file_backup(self.core.ignored_assets_filename_log)
        self.create_file_backup(self.core.notfound_assets_filename_log)
        self.create_file_backup(self.core.bad_data_assets_filename_log)

    def create_asset_from_data(
        self, item, asset_id: str, no_text_data: str, no_int_data: int, no_float_data: float, bool_true_data: bool, bool_false_data: bool
    ) -> (str, dict):
        """
        Create a dict containing all the data for an asset
        :param item: item to get data from
        :param asset_id: id of the asset
        :param no_text_data: text to use if no text data is found
        :param no_int_data: int value to use if no int data is found
        :param no_float_data: float value to use if no float data is found
        :param bool_true_data: bool (True) value to use if no bool data is found
        :param bool_false_data: bool (False) value to use if no bool data is found
        :return: (asset_id, dict containing all the data for an asset)
        """
        record = {}
        metadata = item.metadata
        uid = metadata['id']
        category = metadata['categories'][0]['path']
        separator = ','
        try:
            tmp_list = [separator.join(item.get('compatibleApps')) for item in metadata['releaseInfo']]
            compatible_versions = separator.join(tmp_list)
        except TypeError as error:
            self.logger.warning(f'Error getting compatibleApps {item.app_name} : {error!r}')
            compatible_versions = no_text_data
        thumbnail_url = ''
        try:
            thumbnail_url = metadata['keyImages'][2]['url']  # 'Image' with 488 height
        except IndexError:
            self.logger.debug(f'asset {item.app_name} has no image')
        date_added = datetime.now().strftime(self.core.default_datetime_format)
        extras_data = None
        try:
            extras_data = self.core.uevmlfs.get_item_extras(item.app_name)
        except AttributeError as error:
            self.logger.warning(f'Error getting extra data for {item.app_name} : {error!r}')
        if extras_data is None:
            extras_data = create_empty_assets_extras(item.app_name)
        asset_url = no_text_data
        review = no_int_data
        price = no_float_data
        purchased = bool_false_data
        discount_price = no_float_data
        supported_versions = no_text_data
        page_title = no_text_data
        grab_result = GrabResult.NO_ERROR.name
        oigin = 'Marketplace'  # by default the asset are from the "MarketPlace"
        on_sale = bool_false_data

        try:
            asset_url = extras_data['asset_url']
            review = extras_data['review']
            price = extras_data['price']
            discount_price = extras_data['discount_price']
            purchased = extras_data['purchased']
            supported_versions = extras_data['supported_versions']
            page_title = extras_data['page_title']
            grab_result = extras_data['grab_result']
            on_sale = extras_data['on_sale']
        except (TypeError, KeyError) as error:
            self.logger.warning(f'Key not found in extra data for {item.app_name} : {error!r}')

        if self.core.engine_version_for_obsolete_assets == '' or supported_versions == '' or supported_versions == no_text_data:
            obsolete = bool_false_data
        else:
            supported_versions_list = supported_versions.lower().replace('UE_', '')
            supported_versions_list = create_list_from_string(supported_versions_list)
            obsolete = bool_true_data
            for _, version in enumerate(supported_versions_list):
                if version == '':
                    continue
                else:
                    if float(self.core.engine_version_for_obsolete_assets) <= float(version):
                        obsolete = bool_false_data
                        break
        try:
            values = (
                # dans les infos
                asset_id  # 'asset_id'
                , item.app_name  # 'App name'
                , item.app_title  # 'App title'
                , category  # 'Category'
                , review  # 'Review'
                , metadata['developer']  # 'Developer'
                , metadata['description']  # 'Description'
                , metadata['status']  # 'status'
                , discount_price  # 'Discount price'
                , on_sale  # 'On Sale'
                , purchased  # 'Purchased'
                , obsolete  # 'obsolete'
                , supported_versions  # 'supported versions'
                , grab_result  # 'Grab result'
                , price  # 'Price'
                , no_float_data  # 'Old price'
                # User Fields
                , no_text_data  # 'Comment'
                , no_float_data  # 'Stars'
                , bool_false_data  # 'Must buy'
                , no_text_data  # 'Test result
                , no_text_data  # 'Installed folder'
                , no_text_data  # 'Alternative'
                , oigin  # 'Origin
                # less important fields
                , page_title  # 'page title'
                , thumbnail_url  # 'Image' with 488 height
                , asset_url  # 'Url'
                , compatible_versions  # compatible_versions
                , date_added  # 'Date added'
                , metadata['creationDate']  # 'Creation date'
                , metadata['lastModifiedDate']  # 'Update date'
                , item.app_version('Windows')  # 'UE version'
                , uid  # 'Uid'
            )
            record = dict(zip(CSV_headings.keys(), values))
        except TypeError:
            self.logger.error(f'Could not create record for {item.app_name}')

        return asset_id, record

    def auth(self, args) -> None:
        """
        Handle authentication
        :param args: options passed to the command
        """
        if args.auth_delete:
            self.core.uevmlfs.invalidate_userdata()
            self.logger.info('User data deleted.')
            return

        try:
            self.logger.info('Testing existing login data if present...')
            if self.core.login():
                self.logger.info(
                    'Stored credentials are still valid, if you wish to switch to a different '
                    'account, run "UEVaultManager auth --delete" and try again.'
                )
                return
        except ValueError:
            pass
        except InvalidCredentialsError:
            self.logger.error('Stored credentials were found but were no longer valid. Continuing with login...')
            self.core.uevmlfs.invalidate_userdata()

        # Force an update check and notice in case there are API changes
        self.core.check_for_updates(force=True)
        self.core.force_show_update = True

        if args.import_egs_auth:
            self.logger.info('Importing login session from the Epic Launcher...')
            try:
                if self.core.auth_import():
                    self.logger.info('Successfully imported login session from EGS!')
                    self.logger.info(f'Now logged in as user "{self.core.uevmlfs.userdata["displayName"]}"')
                    return
                else:
                    self.logger.warning('Login session from EGS seems to no longer be valid.')
                    self.core.clean_exit(1)
            except Exception as error:
                self.logger.error(f'No EGS login session found, please login manually. (Exception: {error!r})')
                self.core.clean_exit(1)

        exchange_token = ''
        auth_code = ''
        if not args.auth_code and not args.session_id:
            # only import here since pywebview import is slow
            from UEVaultManager.utils.webview_login import webview_available, do_webview_login

            if not webview_available or args.no_webview or self.core.webview_killswitch:
                # unfortunately the captcha stuff makes a complete CLI login flow kinda impossible right now...
                print('Please login via the epic web login!')
                url = 'https://legendary.gl/epiclogin'
                webbrowser.open(url)
                print(f'If the web page did not open automatically, please manually open the following URL: {url}')
                auth_code = input('Please enter the "authorizationCode" value from the JSON response: ')
                auth_code = auth_code.strip()
                if auth_code[0] == '{':
                    tmp = json.loads(auth_code)
                    auth_code = tmp['authorizationCode']
                else:
                    auth_code = auth_code.strip('"')
            else:
                if do_webview_login(callback_code=self.core.auth_ex_token):
                    self.logger.info(f'Successfully logged in as "{self.core.uevmlfs.userdata["displayName"]}" via WebView')
                else:
                    self.logger.error('WebView login attempt failed, please see log for details.')
                return
        elif args.session_id:
            exchange_token = self.core.auth_sid(args.session_id)
        elif args.auth_code:
            auth_code = args.auth_code
        elif args.ex_token:
            exchange_token = args.ex_token

        if not exchange_token and not auth_code:
            self.logger.fatal('No exchange token/authorization code, cannot login.')
            return

        if exchange_token and self.core.auth_ex_token(exchange_token):
            self.logger.info(f'Successfully logged in as "{self.core.uevmlfs.userdata["displayName"]}"')
        elif auth_code and self.core.auth_code(auth_code):
            self.logger.info(f'Successfully logged in as "{self.core.uevmlfs.userdata["displayName"]}"')
        else:
            self.logger.error('Login attempt failed, please see log for details.')

    def list_assets(self, args) -> None:
        """
        List assets in the vault
        :param args: options passed to the command
        """

        def update_and_merge_csv_record_data(_asset, _items_in_file, _no_data_value) -> []:
            """
            Updates the data of the asset with the data from the items in the file
            :param _asset: asset to update
            :param _items_in_file: list of items in the file
            :param _no_data_value: value to use when no data is available
            :return: list of values to be written in the CSV file
            """
            _asset_id = _asset[0]
            _csv_record = list(_asset[1].values())  # we need a list for the CSV comparison, not a dict
            # merge data from the items in the file (if exists) and those get by the application
            # items_in_file must be a dict of dicts
            if _items_in_file.get(_asset_id):
                item_in_file = _items_in_file.get(_asset_id)
                if len(item_in_file.keys()) != len(CSV_headings.keys()):
                    self.logger.error(
                        f'In the existing file, asset {_asset_id} has not the same number of keys as the CSV headings. This asset is ignored and its values will be overwritten'
                    )
                    return _csv_record
                else:
                    # loops through its columns
                    index = 0
                    price_index = 0
                    _price = float(_no_data_value)
                    old_price = float(_no_data_value)
                    for key, keep_value_in_file in CSV_headings.items():
                        if item_in_file.get(key, None) is None:
                            self.logger.error(
                                f'In the existing file, asset {_asset_id} has no column named {key}. This asset is ignored and its values will be overwritten'
                            )
                            # print(f' CHECK for asset {_asset_id}')
                            return _csv_record
                        else:
                            if keep_value_in_file:
                                value = item_in_file[key]
                                if str_is_bool(value):
                                    value = str_to_bool(value)
                                _csv_record[index] = value
                            # Get the old price in the previous file
                            if key == 'Price':
                                price_index = index
                                try:
                                    _price = float(_csv_record[price_index])
                                    old_price = float(
                                        item_in_file[key]
                                    )  # NOTE: the 'old price' is the 'price' saved in the file, not the 'old_price' in the file
                                except Exception as _error:
                                    self.logger.warning(f'Old price value can not be converted for asset {_asset_id}\nError:{_error!r}')
                            if key == 'Origin':
                                # all the folders when the asset came from are stored in a comma separated list
                                folder_list = item_in_file[key].split(',')
                                # add the new folder to the list without duplicates
                                if _csv_record[index] not in folder_list:
                                    folder_list.append(_csv_record[index])
                                # update the list in the CSV record
                                _csv_record[index] = ','.join(folder_list)
                        index += 1

                _csv_record[price_index + 1] = old_price
            return _csv_record

        def update_and_merge_json_record_data(_asset, _items_in_file, _no_float_value: float, _no_bool_false_value: bool) -> dict:
            """
            Updates the data of the asset with the data from the items in the file
            :param _asset: asset to update
            :param _items_in_file: list of items in the file
            :param _no_float_value:  value to use when no float data is available
            :param _no_bool_false_value: value (False) to use when no bool data is available
            :return:
            """
            _asset_id = _asset[0]
            _json_record = _asset[1]

            # merge data from the items in the file (if exists) and those get by the application
            # items_in_file is a dict of dict
            if _items_in_file.get(_asset_id):
                # loops through its columns
                _price = float(_no_float_value)
                old_price = float(_no_float_value)
                for key, keep_value_in_file in CSV_headings.items():
                    if keep_value_in_file and _items_in_file[_asset_id].get(key):
                        _json_record[key] = _items_in_file[_asset_id][key]

                # Get the old price in the previous file
                try:
                    _price = float(_json_record['Price'])
                    old_price = float(
                        _items_in_file[_asset_id]['Price']
                    )  # NOTE: the 'old price' is the 'price' saved in the file, not the 'old_price' in the file
                except Exception as _error:
                    self.logger.warning(f'Old price values can not be converted for asset {_asset_id}\nError:{_error!r}')
                _json_record['Old price'] = old_price
            return _json_record

        if self.core.create_log_backup:
            self.create_log_file_backup()

        # open log file for assets if necessary
        self.core.setup_assets_logging()
        self.core.egs.notfound_logger = self.core.notfound_logger
        self.core.egs.ignored_logger = self.core.ignored_logger

        output = sys.stdout  # by default, we output to sys.stdout

        self.logger.info('Logging in...')
        if not self.core.login():
            self.logger.error('Login failed, cannot continue!')
            self.core.clean_exit(1)

        if args.force_refresh:
            self.logger.info(
                'force_refresh option is active ...\nRefreshing asset list, this will take several minutes to acheived depending on the internet connection...'
            )
        else:
            self.logger.info('Getting asset list... (this may take a while)')

        if args.filter_category and args.filter_category != gui_g.s.default_category_for_all:
            gui_g.UEVM_filter_category = args.filter_category
            self.logger.info(f'The String "{args.filter_category}" will be search in Assets category')

        gui_g.progress_window_ref = None
        progress_window = None
        if args.gui:
            uewm_gui_exists, progress_window = init_progress_window(args=args, logger=self.logger, callback=self.core.get_asset_list)
            if uewm_gui_exists:
                # if the main gui is running, we already have a tk.mainloop running
                # we need to constantly update the progress bar
                while not progress_window.must_end:
                    progress_window.update()
            else:
                # if the main gui is not running, we need to start a tk.mainloop
                progress_window.mainloop()
            items = progress_window.get_result()

        else:
            items = self.core.get_asset_list(platform='Windows', filter_category=args.filter_category, force_refresh=args.force_refresh)

        if args.include_non_asset:
            na_items = self.core.get_non_asset_library_items(skip_ue=False)
            items.extend(na_items)

        no_int_data = 0
        no_float_value = 0.0
        no_text_data = ''
        no_bool_true_data = True
        no_bool_false_data = False

        # sort assets by name in reverse (to have the latest version first
        items = sorted(items, key=lambda x: x.app_name.lower(), reverse=True)

        # use a dict to store records and avoid duplicates (asset with the same asset_id and a different asset_name)
        assets_to_output = {}
        # create a minimal and full dict of data from existing assets
        assets_in_file = {}

        cpt = 0
        cpt_max = len(items)
        if gui_g.progress_window_ref is not None:
            gui_g.progress_window_ref.reset(new_value=0, new_text="Merging assets data...", new_max_value=len(items))
        for item in items:
            if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                return
            cpt += 1
            # notes:
            #   asset_id is not unique because somme assets can have the same asset_id but with several UE versions
            #   app_name is unique because it includes the unreal version
            #   we use asset_id as key because we don't want to have several entries for the same asset
            #   some asset won't have asset_infos (mainly when using the -T option), in that case we use the app_title as asset_id
            if item.asset_infos.get('Windows'):
                asset_id = item.asset_infos['Windows'].asset_id
            else:
                asset_id = item.app_title

            if assets_to_output.get(asset_id):
                self.logger.debug(f'Asset {asset_id} already present in the list (usually with another ue version)')
            else:
                asset_id, asset = self.create_asset_from_data(
                    item, asset_id, no_text_data, no_int_data, no_float_value, no_bool_true_data, no_bool_false_data
                )  # asset is a dict
                assets_to_output[asset_id] = asset

            if self.core.verbose_mode:
                self.logger.info(f'Asset id={asset_id} has been created from data. Done {cpt}/{cpt_max} items')

        # output with extended info
        if args.output and (args.csv or args.tsv or args.json) and self.core.create_output_backup:
            self.create_file_backup(args.output)

        if args.csv or args.tsv:
            if args.output:
                file_src = args.output
                # If the output file exists, we read its content to keep some data
                try:
                    with open(file_src, 'r', encoding='utf-8') as output:
                        csv_file_content = csv.DictReader(output)
                        # get the data (it's a dict)
                        for csv_record in csv_file_content:
                            # noinspection PyTypeChecker
                            asset_id = csv_record['Asset_id']
                            assets_in_file[asset_id] = csv_record
                        output.close()
                except (FileExistsError, OSError, UnicodeDecodeError, StopIteration):
                    self.logger.warning(f'Could not read CSV record from the file {args.output}')

                # write the content of the file to keep some data
                if not check_and_create_path(file_src):
                    self.logger.critical(f'Could not create folder for {file_src}')
                    self.core.clean_exit(1)
                # reopen file for writing
                output = open(file_src, 'w', encoding='utf-8')
            # end if args.output:

            try:
                writer = csv.writer(output, dialect='excel-tab' if args.tsv else 'excel', lineterminator='\n')
                writer.writerow(CSV_headings.keys())
                cpt = 0
                if gui_g.progress_window_ref is not None:
                    gui_g.progress_window_ref.reset(
                        new_value=0, new_text="Writing assets into csv file...", new_max_value=len(assets_to_output.items())
                    )
                for asset in sorted(assets_to_output.items()):
                    if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                        return
                    asset_id = asset[0]
                    try:
                        if len(assets_in_file) > 0:
                            csv_record_merged = update_and_merge_csv_record_data(asset, assets_in_file, no_int_data)
                        else:
                            csv_record_merged = list(asset[1].values())
                        cpt += 1
                        writer.writerow(csv_record_merged)
                    except (OSError, UnicodeEncodeError, TypeError) as error:
                        self.logger.error(f'Could not write CSV record for {asset_id} into {args.output}\nError:{error!r}')
            except OSError:
                self.logger.error(f'Could not write list result to {args.output}')
        # end if args.csv or args.tsv:

        if args.json:
            if args.output:
                file_src = args.output
                # If the output file exists, we read its content to keep some data
                try:
                    with open(file_src, 'r', encoding='utf-8') as output:
                        assets_in_file = json.load(output)
                        output.close()
                except (FileExistsError, OSError, UnicodeDecodeError, StopIteration, json.decoder.JSONDecodeError):
                    self.logger.warning(f'Could not read Json record from the file {args.output}')

                # write the content of the file to keep some data
                if not check_and_create_path(file_src):
                    self.logger.critical(f'Could not create folder for {file_src}')
                    self.core.clean_exit(1)
                # reopen file for writing
                output = open(file_src, 'w', encoding='utf-8')
            # end if args.output:

            try:
                cpt = 0
                json_content = {}
                if gui_g.progress_window_ref is not None:
                    gui_g.progress_window_ref.reset(
                        new_value=0, new_text="Writing assets into json file...", new_max_value=len(assets_to_output.items())
                    )
                for asset in sorted(assets_to_output.items()):
                    if gui_g.progress_window_ref is not None and not gui_g.progress_window_ref.update_and_continue(increment=1):
                        return
                    asset_id = asset[0]
                    if len(assets_in_file) > 0:
                        json_record_merged = update_and_merge_json_record_data(asset, assets_in_file, no_float_value, no_bool_false_data)
                    else:
                        json_record_merged = asset[1]
                    #      output.write(",\n")
                    try:
                        asset_id = json_record_merged['Asset_id']
                        json_content[asset_id] = json_record_merged
                        cpt += 1
                    except (OSError, UnicodeEncodeError, TypeError) as error:
                        self.logger.error(f'Could not write Json record for {asset_id} into {args.output}\nError:{error!r}')

                json.dump(json_content, output, indent=2)
            except OSError:
                self.logger.error(f'Could not write list result to {args.output}')
            # end if args.json:

        if args.csv or args.tsv or args.json or args.input or args.gui:
            # close the opened file
            if output is not None:
                output.close()
            self.logger.info(
                f'\n======\n{cpt} assets have been printed or saved (without duplicates due to different UE versions)\nOperation Finished\n======\n'
            )
            if args.gui:
                # During the editabletable initial rebuild_data process, the window will not close
                # So we try to close it several times
                max_tries = 3
                progress_window.quit_on_close = True  # gentle quit
                tries = 0
                while progress_window is not None and progress_window.winfo_viewable() and tries < max_tries:
                    progress_window.close_window()
                    time.sleep(0.2)
                    tries += 1
                progress_window.quit_on_close = False  # force destroy the window
                progress_window.close_window()
            return

        # here, no other output has been done before, so we print the asset in a quick format to the console
        print('\nAvailable UE Assets:')
        for asset in items:
            version = asset.app_version('Windows')
            print(f' * {asset.app_title.strip()} (App name: {asset.app_name} | Version: {version})')
        print(f'\nTotal: {len(items)}')

    def list_files(self, args):
        """
        List files for a given app name or manifest url/path
        :param args: options passed to the command
        :return:
        """
        if not args.override_manifest and not args.app_name:
            print('You must provide either a manifest url/path or app name!')
            return
        elif args.app_name:
            args.app_name = self._resolve_aliases(args.app_name)

        # check if we even need to log in
        if args.override_manifest:
            self.logger.info(f'Loading manifest from "{args.override_manifest}"')
            manifest_data, _ = self.core.get_uri_manifest(args.override_manifest)
        else:
            self.logger.info(f'Logging in and downloading manifest for {args.app_name}')
            if not self.core.login():
                self.logger.error('Login failed! Cannot continue with download process.')
                self.core.clean_exit(1)
            update_meta = args.force_refresh
            item = self.core.get_item(args.app_name, update_meta=update_meta)
            if not item:
                self.logger.critical(f'Could not fetch metadata for "{args.app_name}" (check spelling/account ownership)')
                self.core.clean_exit(1)
            manifest_data, _ = self.core.get_cdn_manifest(item, platform='Windows')

        manifest = self.core.load_manifest(manifest_data)
        files = sorted(manifest.file_manifest_list.elements, key=lambda a: a.filename.lower())

        content = ''
        if args.hashlist:
            for fm in files:
                content += f'{fm.hash.hex()} *{fm.filename}\n'
        elif args.csv or args.tsv:
            writer = csv.writer(sys.stdout, dialect='excel-tab' if args.tsv else 'excel', lineterminator='\n')
            writer.writerow(['path', 'hash', 'size', 'install_tags'])
            writer.writerows((fm.filename, fm.hash.hex(), fm.file_size, '|'.join(fm.install_tags)) for fm in files)
        elif args.json:
            _files = [
                dict(filename=fm.filename, sha_hash=fm.hash.hex(), install_tags=fm.install_tags, file_size=fm.file_size, flags=fm.flags)
                for fm in files
            ]
            print(content)
            return self._print_json(_files, args.pretty_json)
        else:
            install_tags = set()
            for fm in files:
                content += fm.filename + '\n'
                for t in fm.install_tags:
                    install_tags.add(t)
            if install_tags:
                # use the log output so this isn't included when piping file list into file
                self.logger.info(f'Install tags: {", ".join(sorted(install_tags))}')

        if args.gui:
            uewm_gui_exists, _ = init_display_window(self.logger)
            custom_print(content, keep_mode=False)  # as it, next print will not keep the content
            if not uewm_gui_exists:
                gui_g.UEVM_gui_ref.mainloop()
        else:
            print(content)

    def status(self, args) -> None:
        """
        Print the information about the vault and the available assets
        :param args: options passed to the command
        """
        if not args.offline:
            try:
                if not self.core.login():
                    self.logger.error('Log in failed!')
                    self.core.clean_exit(1)
            except ValueError:
                pass
            # if automatic checks are off force an update here
            self.core.check_for_updates(force=True)

        if not self.core.uevmlfs.userdata:
            user_name = '<not logged in>'
            args.offline = True
        else:
            user_name = self.core.uevmlfs.userdata['displayName']

        cache_information = self.core.uevmlfs.get_ue_assets_cache_data()
        update_information = self.core.uevmlfs.get_cached_version()
        last_update = update_information.get('last_update', '')
        update_information = update_information.get('data', None)
        last_cache_update = cache_information.get('last_update', '')
        if last_update != '':
            last_update = time.strftime("%x", time.localtime(last_update))
        if last_cache_update != '':
            last_cache_update = time.strftime("%x", time.localtime(last_cache_update))

        json_content = {
            'Epic account': user_name,  #
            'Last data update': last_update,
            'Last cache update': last_cache_update,
            'Config directory': self.core.uevmlfs.path,
            'Platform': f'{platform()} ({os.name})',
            'Current version': f'{UEVM_version} - {UEVM_codename}',
        }
        if not args.offline:
            assets_available = len(self.core.get_asset_list(update_assets=args.force_refresh))
            json_content['Assets available'] = assets_available
            json_content['Update available'] = 'yes' if self.core.update_available else 'no'

            if self.core.update_available and update_information is not None:
                json_content['Update info'] = '\n'
                json_content['New version'] = f'{update_information["version"]} - {update_information["codename"]}'
                json_content['Release summary'] = update_information['summary']
                json_content['Release Url'] = update_information['release_url']
                json_content['Update recommendation'] = update_information['severity']

        if args.json:
            return self._print_json(json_content, args.pretty_json)

        if args.gui:
            uewm_gui_exists, _ = init_display_window(self.logger)
            json_print_key_val(json_content, output_on_gui=True)
            if not uewm_gui_exists:
                gui_g.UEVM_gui_ref.mainloop()
        else:
            json_print_key_val(json_content)

        # prevent update message on close
        self.core.update_available = False

    def info(self, args) -> None:
        """
        Print information about a given app name or manifest url/path
        :param args: options passed to the command
        """
        name_or_path = args.app_name_or_manifest or args.app_name
        app_name = manifest_uri = None
        if os.path.exists(name_or_path) or name_or_path.startswith('http'):
            manifest_uri = name_or_path
        else:
            app_name = self._resolve_aliases(name_or_path)

        if not args.offline and not manifest_uri:
            try:
                if not self.core.login():
                    self.logger.error('Log in failed!')
                    self.core.clean_exit(1)
            except ValueError:
                pass

        # lists that will be printed or turned into JSON data
        info_items = dict(assets=list(), manifest=list(), install=list())
        InfoItem = namedtuple('InfoItem', ['name', 'json_name', 'value', 'json_value'])

        update_meta = not args.offline and args.force_refresh

        item = self.core.get_item(app_name, update_meta=update_meta, platform='Windows')
        if item and not self.core.asset_available(item, platform='Windows'):
            self.logger.warning(
                f'Asset information for "{item.app_name}" is missing, this may be due to the asset '
                f'not being available on the selected platform or currently logged-in account.'
            )
            args.offline = True

        manifest_data = None
        # entitlements = None
        # load installed manifest or URI
        if args.offline or manifest_uri:
            if manifest_uri and manifest_uri.startswith('http'):
                r = self.core.egs.unauth_session.get(manifest_uri)
                r.raise_for_status()
                manifest_data = r.content
            elif manifest_uri and os.path.exists(manifest_uri):
                with open(manifest_uri, 'rb') as f:
                    manifest_data = f.read()
            else:
                self.logger.info('Asset not installed and offline mode enabled, cannot load manifest.')
        elif item:
            # entitlements = self.core.egs.get_user_entitlements()
            egl_meta, status_code = self.core.egs.get_item_info(item.namespace, item.catalog_item_id)
            if status_code != 200:
                self.logger.warning(f'Failed to fetch metadata for {item.app_name}: reponse code = {status_code}')
            item.metadata = egl_meta
            # Get manifest if asset exists for current platform
            if 'Windows' in item.asset_infos:
                manifest_data, _ = self.core.get_cdn_manifest(item, 'Windows')

        if item:
            asset_infos = info_items['assets']
            asset_infos.append(InfoItem('App name', 'app_name', item.app_name, item.app_name))
            asset_infos.append(InfoItem('Title', 'title', item.app_title, item.app_title))
            asset_infos.append(InfoItem('Latest version', 'version', item.app_version('Windows'), item.app_version('Windows')))
            all_versions = {k: v.build_version for k, v in item.asset_infos.items()}
            asset_infos.append(InfoItem('All versions', 'platform_versions', all_versions, all_versions))

        if manifest_data:
            manifest_info = info_items['manifest']
            manifest = self.core.load_manifest(manifest_data)
            manifest_size = len(manifest_data)
            manifest_size_human = f'{manifest_size / 1024:.01f} KiB'
            manifest_info.append(InfoItem('Manifest size', 'size', manifest_size_human, manifest_size))
            manifest_type = 'JSON' if hasattr(manifest, 'json_data') else 'Binary'
            manifest_info.append(InfoItem('Manifest type', 'type', manifest_type, manifest_type.lower()))
            manifest_info.append(InfoItem('Manifest version', 'version', manifest.version, manifest.version))
            manifest_info.append(InfoItem('Manifest feature level', 'feature_level', manifest.meta.feature_level, manifest.meta.feature_level))
            manifest_info.append(InfoItem('Manifest app name', 'app_name', manifest.meta.app_name, manifest.meta.app_name))
            manifest_info.append(InfoItem('Launch EXE', 'launch_exe', manifest.meta.launch_exe or 'N/A', manifest.meta.launch_exe))
            manifest_info.append(InfoItem('Launch Command', 'launch_command', manifest.meta.launch_command or '(None)', manifest.meta.launch_command))
            manifest_info.append(InfoItem('Build version', 'build_version', manifest.meta.build_version, manifest.meta.build_version))
            manifest_info.append(InfoItem('Build ID', 'build_id', manifest.meta.build_id, manifest.meta.build_id))
            if manifest.meta.prereq_ids:
                human_list = [
                    f'Prerequisite IDs: {", ".join(manifest.meta.prereq_ids)}', f'Prerequisite name: {manifest.meta.prereq_name}',
                    f'Prerequisite path: {manifest.meta.prereq_path}', f'Prerequisite args: {manifest.meta.prereq_args or "(None)"}',
                ]
                manifest_info.append(
                    InfoItem(
                        'Prerequisites', 'prerequisites', human_list,
                        dict(
                            ids=manifest.meta.prereq_ids,
                            name=manifest.meta.prereq_name,
                            path=manifest.meta.prereq_path,
                            args=manifest.meta.prereq_args
                        )
                    )
                )
            else:
                manifest_info.append(InfoItem('Prerequisites', 'prerequisites', None, None))

            install_tags = {''}
            for fm in manifest.file_manifest_list.elements:
                for tag in fm.install_tags:
                    install_tags.add(tag)

            install_tags = sorted(install_tags)
            install_tags_human = ', '.join(i if i else '(empty)' for i in install_tags)
            manifest_info.append(InfoItem('Install tags', 'install_tags', install_tags_human, install_tags))
            # file and chunk count
            manifest_info.append(InfoItem('Files', 'num_files', manifest.file_manifest_list.count, manifest.file_manifest_list.count))
            manifest_info.append(InfoItem('Chunks', 'num_chunks', manifest.chunk_data_list.count, manifest.chunk_data_list.count))
            # total file size
            total_size = sum(fm.file_size for fm in manifest.file_manifest_list.elements)
            file_size = '{:.02f} GiB'.format(total_size / 1024 / 1024 / 1024)
            manifest_info.append(InfoItem('Disk size (uncompressed)', 'disk_size', file_size, total_size))
            # total chunk size
            total_size = sum(c.file_size for c in manifest.chunk_data_list.elements)
            chunk_size = '{:.02f} GiB'.format(total_size / 1024 / 1024 / 1024)
            manifest_info.append(InfoItem('Download size (compressed)', 'download_size', chunk_size, total_size))

            # if there are install tags break downsize by tag
            tag_disk_size = []
            tag_disk_size_human = []
            tag_download_size = []
            tag_download_size_human = []
            if len(install_tags) > 1:
                longest_tag = max(max(len(t) for t in install_tags), len('(empty)'))
                for tag in install_tags:
                    # sum up all file sizes for the tag
                    human_tag = tag or '(empty)'
                    tag_files = [fm for fm in manifest.file_manifest_list.elements if (tag in fm.install_tags) or (not tag and not fm.install_tags)]
                    tag_file_size = sum(fm.file_size for fm in tag_files)
                    tag_disk_size.append(dict(tag=tag, size=tag_file_size, count=len(tag_files)))
                    tag_file_size_human = '{:.02f} GiB'.format(tag_file_size / 1024 / 1024 / 1024)
                    tag_disk_size_human.append(f'{human_tag.ljust(longest_tag)} - {tag_file_size_human} '
                                               f'(Files: {len(tag_files)})')
                    # tag_disk_size_human.append(f'Size: {tag_file_size_human}, Files: {len(tag_files)}, Tag: "{tag}"')
                    # accumulate chunk guids used for this tag and count their size too
                    tag_chunk_guids = set()
                    for fm in tag_files:
                        for cp in fm.chunk_parts:
                            tag_chunk_guids.add(cp.guid_num)

                    tag_chunk_size = sum(c.file_size for c in manifest.chunk_data_list.elements if c.guid_num in tag_chunk_guids)
                    tag_download_size.append(dict(tag=tag, size=tag_chunk_size, count=len(tag_chunk_guids)))
                    tag_chunk_size_human = '{:.02f} GiB'.format(tag_chunk_size / 1024 / 1024 / 1024)
                    tag_download_size_human.append(f'{human_tag.ljust(longest_tag)} - {tag_chunk_size_human} '
                                                   f'(Chunks: {len(tag_chunk_guids)})')

            manifest_info.append(InfoItem('Disk size by install tag', 'tag_disk_size', tag_disk_size_human or 'N/A', tag_disk_size))
            manifest_info.append(InfoItem('Download size by install tag', 'tag_download_size', tag_download_size_human or 'N/A', tag_download_size))

        if not args.json:

            def print_info_item(local_item: InfoItem) -> None:
                """
                Prints an info item to the console
                :param local_item:  The info item to print
                """
                if local_item.value is None:
                    custom_print(f'- {local_item.name}: (None)')
                elif isinstance(local_item.value, list):
                    custom_print(f'- {local_item.name}:')
                    for list_item in local_item.value:
                        custom_print(' + ', list_item)
                elif isinstance(local_item.value, dict):
                    custom_print(f'- {local_item.name}:')
                    for k, v in local_item.value.items():
                        custom_print(f' + {k} : {v}')
                else:
                    custom_print(f'- {local_item.name}: {local_item.value}')

            if args.gui:
                uewm_gui_exists, _ = init_display_window(self.logger)
            else:
                uewm_gui_exists = False

            if info_items.get('asset'):
                custom_print('\nAsset Information:')
                for info_item in info_items['asset']:
                    print_info_item(info_item)
            if info_items.get('install'):
                custom_print('\nInstallation information:')
                for info_item in info_items['install']:
                    print_info_item(info_item)
            if info_items.get('manifest'):
                custom_print('\nManifest information:')
                for info_item in info_items['manifest']:
                    print_info_item(info_item)

            if not any(info_items.values()):
                custom_print('No asset information available.')
            custom_print(keep_mode=False)  # as it, next print will not keep the content
            if args.gui and not uewm_gui_exists:
                gui_g.UEVM_gui_ref.mainloop()
        else:
            json_out = dict(asset=dict(), install=dict(), manifest=dict())
            if info_items.get('asset'):
                for info_item in info_items['asset']:
                    json_out['asset'][info_item.json_name] = info_item.json_value
            if info_items.get('install'):
                for info_item in info_items['install']:
                    json_out['install'][info_item.json_name] = info_item.json_value
            if info_items.get('manifest'):
                for info_item in info_items['manifest']:
                    json_out['manifest'][info_item.json_name] = info_item.json_value
            # set empty items to null
            for key, value in json_out.items():
                if not value:
                    json_out[key] = None
            return self._print_json(json_out, args.pretty_json)

    def cleanup(self, args) -> None:
        """
        Cleans up the local assets data folders and logs
        :param args: options passed to the command
        """
        before = self.core.uevmlfs.get_dir_size()

        # delete metadata
        if args.delete_metadata:
            self.logger.debug('Removing app metadata...')
            self.core.uevmlfs.clean_metadata()

        # delete extras data
        if args.delete_extras_data:
            self.logger.debug('Removing app extras data...')
            self.core.uevmlfs.clean_extras()

        # delete log and backup
        self.logger.debug('Removing logs and backups...')
        self.core.uevmlfs.clean_logs_and_backups()

        self.logger.debug('Removing manifests...')
        self.core.uevmlfs.clean_manifests()

        self.logger.debug('Removing tmp data')
        self.core.uevmlfs.clean_tmp_data()

        after = self.core.uevmlfs.get_dir_size()
        self.logger.info(f'Cleanup complete! Removed {(before - after) / 1024 / 1024:.02f} MiB.')

    def get_token(self, args) -> None:
        """
        Gets the access token for the current user
        :param args: options passed to the command
        """
        if not self.core.login(force_refresh=args.bearer):
            self.logger.error('Login failed!')
            return

        if args.bearer:
            args.json = True
            token = dict(
                token_type='bearer',
                access_token=self.core.egs.user['access_token'],
                expires_in=self.core.egs.user['expires_in'],
                expires_at=self.core.egs.user['expires_at'],
                account_id=self.core.egs.user['account_id']
            )
        else:
            token = self.core.egs.get_asset_token()

        if args.json:
            if args.pretty_json:
                print(json.dumps(token, indent=2, sort_keys=True))
            else:
                print(json.dumps(token))
            return
        self.logger.info(f'Exchange code: {token["code"]}')

    def edit_assets(self, args) -> None:
        """
        Edit assets in the database using a GUI
        :param args: options passed to the command
        """
        if not args.input:
            input_filename = gui_g.s.csv_filename
            self.logger.warning('The file to read data from has not been precised by the --input command option. The default file name will be used.')
        else:
            input_filename = gui_fn.path_from_relative_to_absolute(args.input)
        input_filename = os.path.normpath(input_filename)

        app_icon_filename = gui_fn.path_from_relative_to_absolute(gui_g.s.app_icon_filename)
        gui_g.UEVM_log_ref = self.logger
        gui_g.UEVM_cli_ref = self

        # set output file name from the input one. Used by the "rebuild file content" button (or rebuild_data method)
        init_gui_args(args, additional_args={'output': input_filename})
        rebuild = False
        if not os.path.isfile(input_filename):
            input_filename = gui_fn.create_empty_file(input_filename)
            rebuild = True

        gui_g.UEVM_gui_ref = UEVMGui(
            title=gui_g.s.app_title,
            width=gui_g.s.app_width,
            height=gui_g.s.app_height,
            icon=app_icon_filename,
            screen_index=0,
            file=input_filename,
            rebuild_data=rebuild
        )
        gui_g.UEVM_gui_ref.mainloop()
        # gui_g.UEVM_gui_ref.quit()

    @staticmethod
    def print_help(args, parser=None, forced=False) -> None:
        """
        Prints the help for the command
        :param args:
        :param parser: command line parser. If not provided, gui_g.UEVM_parser_ref will be used
        :param forced: if True, the help will be printed even if the --help option is not present
        """
        if parser is None:
            parser = gui_g.UEVM_parser_ref
        if parser is None:
            return
        uewm_gui_exists = False

        if args.full_help or forced:
            if args.gui:
                uewm_gui_exists, _ = init_display_window()
            custom_print(keep_mode=False, text=parser.format_help())

            # Commands that should not be shown in full help/list of commands (e.g. aliases)
            _hidden_commands = {'download', 'update', 'repair', 'get-token', 'verify-asset', 'list-assets'}
            # Print the help for all the subparsers. Thanks stackoverflow!
            custom_print(text='Individual command help:')
            # noinspection PyProtectedMember,PyUnresolvedReferences
            subparsers = next(a for a in parser._actions if isinstance(a, argparse._SubParsersAction))
            # noinspection PyUnresolvedReferences
            for choice, subparser in subparsers.choices.items():
                if choice in _hidden_commands:
                    continue
                custom_print(text=f'\nCommand: {choice}')
                custom_print(text=subparser.format_help())
        elif os.name == 'nt':
            from UEVaultManager.lfs.windows_helpers import double_clicked
            if double_clicked():
                custom_print(text='Please note that this is not the intended way to run UEVaultManager.')
                custom_print(text='If you want to start it without arguments, you can start it in edit mode by default.')
                custom_print(text='For that, you must set the line start_in_edit_mode=true in the configuration file.')
                custom_print(text='More info on usage and configuration can be found in https://github.com/LaurentOngaro/UEVaultManager#readme')
                subprocess.Popen(['cmd', '/K', 'echo>nul'])
        else:
            # on non-windows systems
            # UEVaultManagerCLI.print_help(args, parser=parser, forced=True)
            UEVaultManagerCLI.print_version()
            return

        if args.gui and not uewm_gui_exists:
            gui_g.UEVM_gui_ref.mainloop()

    @staticmethod
    def print_version():
        """
        Prints the version of UEVaultManager and exit
        """
        print(f'UEVaultManager version "{UEVM_version}", codename "{UEVM_codename}"')
        sys.exit(0)


def main():
    """
    Main function
    """
    parser = argparse.ArgumentParser(description=f'UEVaultManager v{UEVM_version} - "{UEVM_codename}"')
    parser.register('action', 'parsers', HiddenAliasSubparsersAction)

    # general arguments
    parser.add_argument('-H', '--full-help', dest='full_help', action='store_true', help='Show full help (including individual command help)')
    parser.add_argument('-d', '--debug', dest='debug', action='store_true', help='Set loglevel to debug')
    # noinspection DuplicatedCode
    parser.add_argument('-y', '--yes', dest='yes', action='store_true', help='Default to yes for all prompts')
    parser.add_argument('-V', '--version', dest='version', action='store_true', help='Print version and exit')
    parser.add_argument(
        '-c', '--config-file', dest='config_file', action='store', metavar='<path/name>', help='Overwrite the default configuration file name to use'
    )
    parser.add_argument('-J', '--pretty-json', dest='pretty_json', action='store_true', help='Pretty-print JSON. Improve readability')
    parser.add_argument(
        '-A',
        '--api-timeout',
        dest='api_timeout',
        action='store',
        type=float,
        default=10,
        metavar='<seconds>',
        help='API HTTP request timeout (default: 10 seconds)'
    )
    parser.add_argument(
        '-g',
        '--gui',
        dest='gui',
        action='store_true',
        help='Display additional informations using gui elements like dialog boxes or progress window'
    )
    gui_g.UEVM_parser_ref = parser

    # all the commands
    subparsers = parser.add_subparsers(title='Commands', dest='subparser_name', metavar='<command>')
    auth_parser = subparsers.add_parser('auth', help='Authenticate with the Epic Games Store')
    clean_parser = subparsers.add_parser('cleanup', help='Remove old temporary, metadata, and manifest files')
    info_parser = subparsers.add_parser('info', help='Prints info about specified app name or manifest')
    list_parser = subparsers.add_parser('list', aliases=('list-assets',), hide_aliases=True, help='List available assets')
    list_files_parser = subparsers.add_parser('list-files', help='List files in manifest')
    status_parser = subparsers.add_parser('status', help='Show UEVaultManager status information')
    edit_parser = subparsers.add_parser('edit', aliases=('edit-assets',), hide_aliases=True, help='Edit the assets list file')

    # hidden commands have no help text
    get_token_parser = subparsers.add_parser('get-token')

    # Positional arguments
    list_files_parser.add_argument('app_name', nargs='?', metavar='<App Name>', help='Name of the app (optional)')
    info_parser.add_argument('app_name_or_manifest', help='App name or manifest path/URI', metavar='<App Name/Manifest URI>')

    # Flags
    auth_parser.add_argument(
        '--import', dest='import_egs_auth', action='store_true', help='Import Epic Games Launcher authentication data (logs out of EGL)'
    )
    auth_parser.add_argument(
        '--code',
        dest='auth_code',
        action='store',
        metavar='<authorization code>',
        help='Use specified authorization code instead of interactive authentication'
    )
    auth_parser.add_argument(
        '--token',
        dest='ex_token',
        action='store',
        metavar='<exchange token>',
        help='Use specified exchange token instead of interactive authentication'
    )
    auth_parser.add_argument(
        '--sid', dest='session_id', action='store', metavar='<session id>', help='Use specified session id instead of interactive authentication'
    )
    auth_parser.add_argument('--delete', dest='auth_delete', action='store_true', help='Remove existing authentication (log out)')
    auth_parser.add_argument('--disable-webview', dest='no_webview', action='store_true', help='Do not use embedded browser for login')

    list_parser.add_argument(
        '-T', '--third-party', dest='include_non_asset', action='store_true', default=False, help='Include assets that are not installable.'
    )
    # noinspection DuplicatedCode
    list_parser.add_argument('--csv', dest='csv', action='store_true', help='Output in in CSV format')
    list_parser.add_argument('--tsv', dest='tsv', action='store_true', help='Output in in TSV format')
    list_parser.add_argument('--json', dest='json', action='store_true', help='Output in in JSON format')
    list_parser.add_argument(
        '-f',
        '--force-refresh',
        dest='force_refresh',
        action='store_true',
        help='Force a refresh of all assets metadata. It could take some time ! If not forced, the cached data will be used'
    )
    list_parser.add_argument(
        '-fc',
        '--filter-category',
        dest='filter_category',
        action='store',
        help='Filter assets by category. Search against the asset category in the marketplace. Search is case insensitive and can be partial'
    )
    list_parser.add_argument(
        '-o', '--output', dest='output', metavar='<path/name>', action='store', help='The file name (with path) where the list should be written'
    )
    list_parser.add_argument(
        '-g',
        '--gui',
        dest='gui',
        action='store_true',
        help='Display additional informations using gui elements like dialog boxes or progress window'
    )

    list_files_parser.add_argument(
        '--manifest', dest='override_manifest', action='store', metavar='<uri>', help='Manifest URL or path to use instead of the CDN one'
    )
    list_files_parser.add_argument('--csv', dest='csv', action='store_true', help='Output in CSV format')
    list_files_parser.add_argument('--tsv', dest='tsv', action='store_true', help='Output in TSV format')
    # noinspection DuplicatedCode
    list_files_parser.add_argument('--json', dest='json', action='store_true', help='Output in JSON format')
    list_files_parser.add_argument(
        '--hashlist', dest='hashlist', action='store_true', help='Output file hash list in hashCheck/sha1sum -c compatible format'
    )
    list_files_parser.add_argument(
        '-f',
        '--force-refresh',
        dest='force_refresh',
        action='store_true',
        help='Force a refresh of all assets metadata. It could take some time ! If not forced, the cached data will be used'
    )
    list_files_parser.add_argument(
        '-g', '--gui', dest='gui', action='store_true', help='Display the output in a windows instead of using the console'
    )

    status_parser.add_argument('--offline', dest='offline', action='store_true', help='Only print offline status information, do not login')
    status_parser.add_argument('--json', dest='json', action='store_true', help='Show status in JSON format')
    status_parser.add_argument(
        '-f',
        '--force-refresh',
        dest='force_refresh',
        action='store_true',
        help='Force a refresh of all assets metadata. It could take some time ! If not forced, the cached data will be used'
    )
    status_parser.add_argument('-g', '--gui', dest='gui', action='store_true', help='Display the output in a windows instead of using the console')

    clean_parser.add_argument(
        '-m,'
        '--delete-metadata', dest='delete_metadata', action='store_true', help='Also delete metadata files. They are kept by default'
    )
    clean_parser.add_argument(
        '-e,'
        '--delete-extras-data', dest='delete_extras_data', action='store_true', help='Also delete extras data files. They are kept by default'
    )
    # noinspection DuplicatedCode
    info_parser.add_argument('--offline', dest='offline', action='store_true', help='Only print info available offline')
    info_parser.add_argument('--json', dest='json', action='store_true', help='Output information in JSON format')
    info_parser.add_argument(
        '-f',
        '--force-refresh',
        dest='force_refresh',
        action='store_true',
        help='Force a refresh of all assets metadata. It could take some time ! If not forced, the cached data will be used'
    )
    info_parser.add_argument('-g', '--gui', dest='gui', action='store_true', help='Display the output in a windows instead of using the console')

    get_token_parser.add_argument('--json', dest='json', action='store_true', help='Output information in JSON format')
    get_token_parser.add_argument('--bearer', dest='bearer', action='store_true', help='Return fresh bearer token rather than an exchange code')

    edit_parser.add_argument(
        '-i', '--input', dest='input', metavar='<path/name>', action='store', help='The file name (with path) where the list should be read from'
    )
    # not use for now
    # edit_parser.add_argument('--csv', dest='csv', action='store_true', help='Input file is in CSV format')
    # edit_parser.add_argument('--tsv', dest='tsv', action='store_true', help='Input file is in TSV format')
    # edit_parser.add_argument('--json', dest='json', action='store_true', help='Input file is in JSON format')

    # Note: this line prints the full help and quit if not other command is available
    args, extra = parser.parse_known_args()

    if args.version:
        cli.print_version()

    cli = UEVaultManagerCLI(override_config=args.config_file, api_timeout=args.api_timeout)

    start_in_edit_mode = str_to_bool(cli.core.uevmlfs.config.get('UEVaultManager', 'start_in_edit_mode', fallback=False))

    if not start_in_edit_mode and (not args.subparser_name or args.full_help):
        cli.print_help(args=args, parser=parser)
        return

    ql = cli.setup_threaded_logging()

    conf_log_level = cli.core.uevmlfs.config.get('UEVaultManager', 'log_level', fallback='info')
    if conf_log_level == 'debug' or args.debug:
        cli.core.verbose_mode = True
        logging.getLogger().setLevel(level=logging.DEBUG)
        # keep requests quiet
        logging.getLogger('requests').setLevel(logging.WARNING)
        logging.getLogger('urllib3').setLevel(logging.WARNING)

    cli.core.create_output_backup = str_to_bool(cli.core.uevmlfs.config.get('UEVaultManager', 'create_output_backup', fallback=True))
    cli.core.create_log_backup = str_to_bool(cli.core.uevmlfs.config.get('UEVaultManager', 'create_log_backup', fallback=True))
    cli.core.verbose_mode = str_to_bool(cli.core.uevmlfs.config.get('UEVaultManager', 'verbose_mode', fallback=False))
    cli.ue_assets_max_cache_duration = int(cli.core.uevmlfs.config.get('UEVaultManager', 'ue_assets_max_cache_duration', fallback=1296000))

    cli.core.ignored_assets_filename_log = cli.core.uevmlfs.config.get('UEVaultManager', 'ignored_assets_filename_log', fallback='')
    cli.core.notfound_assets_filename_log = cli.core.uevmlfs.config.get('UEVaultManager', 'notfound_assets_filename_log', fallback='')
    cli.core.bad_data_assets_filename_log = cli.core.uevmlfs.config.get('UEVaultManager', 'bad_data_assets_filename_log', fallback='')

    cli.core.engine_version_for_obsolete_assets = cli.core.uevmlfs.config.get('UEVaultManager', 'engine_version_for_obsolete_assets', fallback='4.26')

    # if --yes is used as part of the subparsers arguments manually set the flag in the main parser.
    if '-y' in extra or '--yes' in extra:
        args.yes = True

    # technically args.func() with set defaults could work (see docs on subparsers)
    # but that would require all funcs to accept args and extra...
    try:
        if args.subparser_name == 'auth':
            cli.auth(args)
        elif args.subparser_name in {'list', 'list-assets'}:
            cli.list_assets(args)
        elif args.subparser_name == 'list-files':
            cli.list_files(args)
        elif args.subparser_name == 'status':
            cli.status(args)
        elif args.subparser_name == 'info':
            cli.info(args)
        elif args.subparser_name == 'cleanup':
            cli.cleanup(args)
        elif args.subparser_name == 'get-token':
            cli.get_token(args)
        elif args.subparser_name in {'edit', 'edit-assets'}:
            cli.edit_assets(args)
        elif start_in_edit_mode:
            args.subparser_name = 'edit'
            args.gui = True
            args.input = init_gui(False)
            cli.edit_assets(args)
    except KeyboardInterrupt:
        cli.logger.info('Command was aborted via KeyboardInterrupt, cleaning up...')

    # Disable the update message if JSON/TSV/CSV outputs are used
    disable_update_message = False
    if hasattr(args, 'json'):
        disable_update_message = args.json
    if not disable_update_message and hasattr(args, 'tsv'):
        disable_update_message = args.tsv
    if not disable_update_message and hasattr(args, 'csv'):
        disable_update_message = args.csv

    # show note if update is available
    if not disable_update_message and cli.core.update_available and cli.core.update_notice_enabled():
        if update_info := cli.core.get_update_info():
            print(f'\nAn update available!')
            print(f'- New version: {update_info["version"]} - "{update_info["codename"]}"')
            print(f'- Release summary:\n{update_info["summary"]}')
            if update_info['severity'] == UpdateSeverity.HIGH.name:
                print('! This update is recommended as it fixes major issues.')
                print(f'\n- Release URL: {update_info["release_url"]}')

    cli.core.clean_exit()
    ql.stop()
    sys.exit(0)


if __name__ == '__main__':
    # required for pyinstaller on Windows, does nothing on other platforms.
    freeze_support()
    main()



# coding=utf-8
"""
Implementation for:
- WidgetType: enum for the widget types
- ExtendedWidget: base class for all widgets in the app
- ExtendedEntry: extended entry widget
- ExtendedText: extended text widget
- ExtendedLabel: extended label widget
- ExtendedCheckButton: extended checkbutton widget
"""

import inspect
import os
import tkinter as tk
from enum import Enum
from tkinter import ttk
from tkinter.font import nametofont

from UEVaultManager.tkgui.modules.functions import log_warning
from UEVaultManager.tkgui.modules.functions_no_deps import path_from_relative_to_absolute


class WidgetType(Enum):
    """
    Enum for the widget types
    """
    ENTRY = 0  # Entry widget
    TEXT = 1  # Text widget
    LABEL = 2  # Label widget
    CHECKBUTTON = 3  # Checkbutton widget


class ExtendedWidget:
    """
    Base class for all widgets in the app.
    :param tag: tag of the widget
    :param row: row of the widget
    :param col: column of the widget
    :return: ExtendedWidget instance
    """

    def __init__(self, tag=None, row=-1, col=-1, default_content=''):
        self.tag = tag
        self.col = col
        self.row = row
        self.default_content = default_content if default_content else self.tag_to_label(tag)
        # can't call this here because the set_content function is specific and overridden in the derived classes
        # self.reset_content()

    @staticmethod
    def _remove_extended_args(kwargs, function_signature) -> None:
        """
        Removes the extended args from the kwargs.
        :param kwargs: args to extract from.
        :param function_signature: function to get the signature from. We can't use a non-static method because the init function will be the derived class's init function.
        """
        init_args = inspect.signature(function_signature)
        for key in init_args.parameters:
            kwargs.pop(key, None)

    @staticmethod
    def _extract_extended_args(kwargs, function_signature) -> dict:
        """
        Extracts the extended args from the kwargs.
        :param kwargs: args to extract from. Note that the kwargs will be modified and the extended args will be removed.
        :param function_signature: function to get the signature from. We can't use a non-static method because the init function will be the derived class's init function.
        :return: dict of extended args
        """
        init_args = inspect.signature(function_signature)
        result = {}
        for key in init_args.parameters:
            if key == 'self':
                continue
            result[key] = kwargs.pop(key, None)
        return result

    @staticmethod
    def tag_to_label(tag: str or None) -> str:
        """
        Convert a tag to a label
        :param tag: the tag to convert
        :return: the label
        """
        if tag is None:
            return ''

        return tag.capitalize().replace('_', ' ')

    def set_content(self, content='') -> None:
        """
        Sets the content of the widget.
        :param content: content to set
        """
        try:
            # noinspection PyUnresolvedReferences
            self.config(text=content)
        except (AttributeError, tk.TclError) as error:
            log_warning(f'Failed to set content of {self} to {content}: {error!r}')

    def get_content(self) -> str:
        """
        Gets the content of the widget.
        :return: content of the widget
        """
        try:
            # noinspection PyUnresolvedReferences
            return self.get()
        except (AttributeError, tk.TclError) as error:
            log_warning(f'Failed to get content of {self}: {error!r}')
            return ''

    def reset_content(self) -> None:
        """
        Resets the content of the widget to the default content.
        """
        self.set_content(self.default_content)

    def get_style(self) -> ttk.Style:
        """
        Get the ttk.Style object of the widget.
        :return: ttk.Style object
        """
        try:
            # noinspection PyUnresolvedReferences
            style = self.winfo_toplevel().style  # only works if the widget is a child of UEVMGui,
        except AttributeError:
            try:
                # noinspection PyProtectedMember
                root = tk._default_root.style
                style_name = root.theme_use()
                style = root.Style(style_name)
            except AttributeError:
                style = None
        return style

    def get_default_font(self) -> tk.font.Font:
        """
        Get the default font for ttk widgets. If the default font is not found, use the TkDefaultFont.
        :return: The default font for ttk widgets.
        """
        default_font = nametofont("TkDefaultFont")
        style = self.get_style()
        if style is not None:
            default_font = style.lookup("TEntry", "font")
            if default_font == '':
                # noinspection PyUnresolvedReferences
                default_font = self.cget("font")

        return default_font


class ExtendedEntry(ExtendedWidget, ttk.Entry):
    """
    Extended widget version of a ttk.Entry class.
    :param master: container for the widget
    :param kwargs: kwargs to pass to the widget
    :return: ExtendedEntry instance
    """

    def __init__(self, master=None, **kwargs):
        if master is None:
            print('A container is needed to display this widget')
            return
        ext_args = self._extract_extended_args(kwargs, function_signature=ExtendedWidget.__init__)
        ExtendedWidget.__init__(self, **ext_args)
        # already made with _extract_extended_args
        # self._remove_extended_args(kwargs, function_signature=ExtendedWidget.__init__)
        ttk.Entry.__init__(self, master, **kwargs)
        self.widget_type = WidgetType.ENTRY

    def set_content(self, content='') -> None:
        """
        Sets the content of the widget.
        :param content: content to set
        """
        self.delete(0, tk.END)
        self.insert(0, content)


class ExtendedText(ExtendedWidget, tk.Text):
    """
    Extended widget version of a ttk.Text. Also add a "ttk.style" like property to the widget.
    :param master: container for the widget
    :param kwargs: kwargs to pass to the widget
    :return: ExtendedText instance
    """

    def __init__(self, master=None, **kwargs):
        if master is None:
            print('A container is needed to display this widget')
            return
        ext_args = self._extract_extended_args(kwargs, function_signature=ExtendedWidget.__init__)
        ExtendedWidget.__init__(self, **ext_args)

        # already made with _extract_extended_args
        # self._remove_extended_args(kwargs, function_signature=ExtendedWidget.__init__)
        tk.Text.__init__(self, master, **kwargs)
        self.widget_type = WidgetType.TEXT
        self.update_style()

    def update_style(self) -> None:
        """
        Update the style of the widget based on a ttk.Entry widget.
        """
        style = self.get_style()

        if style is None:
            return
        # get some style from the ttk.Entry widget
        font = self.get_default_font()
        bg_color = style.lookup('TEntry', 'fieldbackground', default='white')
        fg_color = style.lookup('TEntry', 'foreground', default='black')
        relief = style.lookup('TEntry', 'relief', default='flat')
        # border_color = style.lookup('TEntry', 'bordercolor', default='black')
        self.configure(background=bg_color, foreground=fg_color, borderwidth=1, relief=relief, font=font)

    def set_content(self, content='') -> None:
        """
        Sets the content of the widget.
        :param content: content to set
        """
        try:
            self.delete('1.0', tk.END)
            self.insert('1.0', content)
        except (AttributeError, tk.TclError) as error:
            log_warning(f'Failed to set content of {self} to {content}: {error!r}')

    def get_content(self) -> str:
        """
        Gets the content of the widget.
        :return: content of the widget
        """
        try:
            return self.get('1.0', tk.END)
        except (AttributeError, tk.TclError) as error:
            log_warning(f'Failed to get content of {self}: {error!r}')
            return ''


class ExtendedLabel(ExtendedWidget, ttk.Label):
    """
    Extended widget version of a ttk.Label.
    :param master: container for the widget
    :param text: Text to display next to the checkbutton
    :param kwargs: kwargs to pass to the widget
    :return: ExtendedLabel instance
    """

    def __init__(self, master=None, **kwargs):
        if master is None:
            print('A container is needed to display this widget')
            return
        ext_args = self._extract_extended_args(kwargs, function_signature=ExtendedWidget.__init__)
        ExtendedWidget.__init__(self, **ext_args)
        # already made with _extract_extended_args
        # self._remove_extended_args(kwargs, function_signature=ExtendedWidget.__init__)
        ttk.Label.__init__(self, master, **kwargs)
        self.widget_type = WidgetType.LABEL


class ExtendedCheckButton(ExtendedWidget):
    """
    Create a new widget version of a ttk.Checkbutton.
    Note: We don't use the ttk.Checkbutton because it's hard to sync its state when using the on_click event.
    :param master: Parent widget
    :param label: Text to display next to the checkbutton
    :param images_folder: Path to the folder containing the images for the checkbutton. If empty, the './assets' folder will be used
    :param change_state_on_click: If True, the state of the checkbutton will change when clicking on the text or the checkbutton. if not, the change must be done manually by calling the switch_state method
    :param kwargs: kwargs to pass to the widget
    :return: ExtendedCheckButton instance
    """

    def __init__(self, master, label=None, images_folder=None, change_state_on_click=True, **kwargs):
        if master is None:
            print('A container is needed to display this widget')
            return
        ext_args = self._extract_extended_args(kwargs, function_signature=ExtendedWidget.__init__)
        ExtendedWidget.__init__(self, **ext_args)
        # by default , images are searched in a folder named 'statics' in the directory of this file
        if images_folder is None:
            images_folder = path_from_relative_to_absolute('./assets/')
        self._img_checked = tk.PhotoImage(file=os.path.join(images_folder, 'checked_16.png'))  # Path to the checked image
        self._img_uncheckked = tk.PhotoImage(file=os.path.join(images_folder, 'unchecked_16.png'))  # Path to the unchecked image
        self.widget_type = WidgetType.CHECKBUTTON
        self.default_content = False
        self._var = tk.BooleanVar(value=self.default_content)
        frm_inner = ttk.Frame(master)
        lbl_text = ttk.Label(frm_inner, text='')  # no text bydefault
        check_label = ttk.Label(frm_inner, image=self._img_uncheckked, cursor='hand2')
        lbl_text.pack(side=tk.LEFT)
        check_label.pack(side=tk.LEFT)
        self._frm_inner = frm_inner
        self._lbl_text = lbl_text
        self._check_label = check_label

        if label is not None:
            self.set_label(label)
        # noinspection PyTypeChecker
        # keep "bad" type to keep compatible signatures with overriden methods
        self.set_content(self.default_content)

        if change_state_on_click:
            self.bind("<Button-1>", self.switch_state)

    def _update_state(self) -> None:
        """
        Updates the image of the checkbutton
        """
        current_state = self._var.get()
        if current_state:
            self._check_label.config(image=self._img_checked)
        else:
            self._check_label.config(image=self._img_uncheckked)

    def pack(self, **kwargs) -> None:
        """
        Packs the widget
        :param kwargs: kwargs to pass to the widget
        """
        self._frm_inner.pack(**kwargs)

    def grid(self, **kwargs) -> None:
        """
        Grids the widget
        :param kwargs: kwargs to pass to the widget
        """
        self._frm_inner.grid(**kwargs)

    def bind(self, sequence=None, command=None) -> None:
        """
        Binds a callback to the widget
        :param sequence: Sequence to bind to
        :param command:  function to bind
        """
        self._lbl_text.bind(sequence, command)
        self._check_label.bind(sequence, command)

    # noinspection PyUnusedLocal
    def switch_state(self, event=None) -> bool:
        """
        Switches the state of the checkbutton
        :param event: event that triggered the call
        """
        value = bool(self._var.get())
        # print(f'Current state: {value} event: {event}   ')
        value = not value
        self._var.set(value)
        self._update_state()
        return value

    def set_label(self, text='') -> None:
        """
        Sets the label of the widget
        :param text: text to set
        """
        self._check_label.config(text=text)

    def set_content(self, content='') -> None:
        """
        Sets the content of the widget. True, 'True' and '1' will be considered as True, everything else will be considered as False
        :param content: content to set
        """
        try:
            if type(content) is bool:
                # noinspection PyTypeChecker
                # keep "bad" type to keep compatible signatures with overriden methods
                self._var.set(content)
            elif (content.lower() == 'true') or (content == '1'):
                self._var.set(True)
            else:
                self._var.set(False)

            self._update_state()
        except (AttributeError, tk.TclError) as error:
            log_warning(f'Failed to set content of {self} to {content}: {error!r}')

    def get_content(self) -> bool:
        """
        Gets the content of the widget.
        :return: True if the checkbutton is checked, False otherwise
        """
        return self._var.get()



# coding=utf-8
"""
Implementation for:
- AppConf: ConfigParser subclass that saves modification time of config file
"""
import configparser
import os
import time


class AppConf(configparser.ConfigParser):
    """
    ConfigParser subclass that saves modification time of config file
    :param args: arguments
    :param kwargs: keyword arguments
    """

    def __init__(self, *args, **kwargs):
        self.modified = False
        self.read_only = False
        self.mod_time = None
        super().__init__(*args, **kwargs)
        self.optionxform = str

    def read(self, filename: str, **kwargs) -> list:
        """
        Read config file and save modification time
        :param filename: file to read
        :param kwargs: keyword arguments
        :return: content of the file
        """
        # if config file exists, save modification time
        if os.path.exists(filename):
            self.mod_time = int(os.stat(filename).st_mtime)

        return super().read(filename)

    def write(self, *args, **kwargs) -> None:
        """
        Write config file and save modification time
        :param args: arguments
        :param kwargs: keyword arguments
        """
        self.modified = False
        super().write(*args, **kwargs)
        self.mod_time = int(time.time())

    def set(self, section: str, option: str, value=None) -> None:
        """
        Set a config option
        :param section: section name in the config file. If the section does not exist, it will be created
        :param option: option name
        :param value: value to set
        """
        if self.read_only:
            return

        # ensure config section exists
        if not self.has_section(section):
            self.add_section(section)

        self.modified = True
        super().set(section, option, value)

    def remove_option(self, section: str, option: str) -> bool:
        """
        Remove an option from the config file
        :param section: section name in the config file. If the section does not exist, it will be created
        :param option: option name
        :return: True if the option was removed, False otherwise
        """
        if self.read_only:
            return False

        self.modified = True
        return super().remove_option(section, option)

    def __setitem__(self, key: str, value) -> None:
        """
        (internal) Set a config option
        :param key: key name
        :param value: value to set
        """
        if self.read_only:
            return

        self.modified = True
        super().__setitem__(key, value)



# coding=utf-8
"""
Implementation for:
- GUISettings: class containing all the settings for the GUI
"""
import os

from termcolor import colored

import UEVaultManager.tkgui.modules.functions_no_deps as gui_fn
from UEVaultManager.lfs.utils import clean_filename
from UEVaultManager.models.config import AppConf

# NOTE : we can't import the following modules here because of circular dependencies
# UEVaultManager.tkgui.modules.functions_no_deps


def log(msg: str) -> None:
    """
    Print a message to the console
    :param msg: the message to log
    """
    msg = colored(msg, 'orange')
    print(msg)


class GUISettings:
    """
    This class contains all the settings for the GUI.
    :param config_file: Path to config file to use instead of default

    """

    def __init__(self, config_file=None):
        self.path = ''
        self.config_path = ''
        self.config = AppConf(comment_prefixes='/', allow_no_value=True)

        self.init_gui_config_file(config_file)

        # ##### start of properties stored in config file
        # store all the properties that must be saved in config file
        self.config_vars = {
            'debug_mode': gui_fn.convert_to_bool(self.config.get('UEVaultManager', 'debug_mode', fallback=False)),
            'never_update_data_files': gui_fn.convert_to_bool(self.config.get('UEVaultManager', 'never_update_data_files', fallback=False)),
            'reopen_last_file': gui_fn.convert_to_bool(self.config.get('UEVaultManager', 'reopen_last_file', fallback=False)),
            'use_colors_for_data': gui_fn.convert_to_bool(self.config.get('UEVaultManager', 'use_colors_for_data', fallback=True)),
            'image_cache_max_time': int(self.config.get('UEVaultManager', 'image_cache_max_time', fallback=60 * 60 * 24 * 15)),
            'last_opened_file': self.config.get('UEVaultManager', 'last_opened_file', fallback=''),
            'cache_folder': self.config.get('UEVaultManager', 'cache_folder', fallback='../../../cache'),
            'results_folder': self.config.get('UEVaultManager', 'results_folder', fallback='../../../results')
        }

        # the following folders are relative to the current file location
        # they must be used trought path_from_relative_to_absolute
        # following vars are not set as properties to avoid storing absolute paths in the config file
        self.cache_folder = gui_fn.path_from_relative_to_absolute(self.config_vars['cache_folder'])
        self.results_folder = gui_fn.path_from_relative_to_absolute(self.config_vars['results_folder'])

        # Folder for assets (aka. images, icon... not "UE assets") used for the GUI. THIS IS NOT A SETTING THAT CAN BE CHANGED BY THE USER
        self.assets_folder = gui_fn.path_from_relative_to_absolute('../../assets')

        self.app_icon_filename = os.path.join(self.assets_folder, 'main.ico')
        self.default_image_filename = os.path.join(self.assets_folder, 'UEVM_200x200.png')

        if self.config_vars['reopen_last_file'] and os.path.isfile((self.config_vars['last_opened_file'])):
            self.csv_filename = self.config_vars['last_opened_file']
        else:
            self.csv_filename = os.path.join(self.config_vars['results_folder'], 'list.csv')

        self.csv_filename = os.path.normpath(self.csv_filename)
        self.app_title = 'UEVM Gui'
        self.app_width = 1600
        self.app_height = 935
        self.app_monitor = 1
        self.csv_datetime_format = '%Y-%m-%d %H:%M:%S'
        self.data_filetypes = (('csv file', '*.csv'), ('tcsv file', '*.tcsv'), ('json file', '*.json'), ('text file', '*.txt'))
        self.preview_max_width = 150
        self.preview_max_height = 150
        self.default_global_search = 'Text to search...'
        self.default_category_for_all = 'All'
        self.empty_cell = 'nan'
        self.expand_columns_factor = 20
        self.contract_columns_factor = 20
        # ttkbootstrap themes:
        # light themes : "cosmo", "flatly", "litera", "minty", "lumen", "sandstone", "yeti", "pulse", "united", "morph", "journal", "simplex", "cerculean"
        # dark themes: "darkly", "superhero", "solar", "cyborg", "vapor"
        self.theme_name = 'lumen'
        self.theme_font = ('Verdana', 8)
        self.datatable_default_pref = {
            'align': 'w',  #
            'cellbackgr': '#F4F4F3',  #
            'cellwidth': 100,  #
            'floatprecision': 2,  #
            'thousandseparator': '',  #
            'font': 'Verdana',  #
            'fontsize': 8,  #
            'fontstyle': '',  #
            'grid_color': '#ABB1AD',  #
            'linewidth': 1,  #
            'rowheight': 22,  #
            'rowselectedcolor': '#E4DED4',  #
            'textcolor': 'black'  #
        }

    def get_debug_mode(self):
        """ Getter for debug_mode """
        return self.config_vars['debug_mode']

    def set_debug_mode(self, value):
        """ Setter for debug_mode """
        self.config_vars['debug_mode'] = value

    def get_never_update_data_files(self):
        """ Getter for never_update_data_files """
        return self.config_vars['never_update_data_files']

    def set_never_update_data_files(self, value):
        """ Setter for never_update_data_files """
        self.config_vars['never_update_data_files'] = value

    def get_reopen_last_file(self):
        """ Getter for reopen_last_file """
        return self.config_vars['reopen_last_file']

    def set_reopen_last_file(self, value):
        """ Setter for reopen_last_file """
        self.config_vars['reopen_last_file'] = value

    def get_use_colors_for_data(self):
        """ Getter for use_colors_for_data """
        return self.config_vars['use_colors_for_data']

    def set_use_colors_for_data(self, value):
        """ Setter for use_colors_for_data """
        self.config_vars['use_colors_for_data'] = value

    def get_image_cache_max_time(self):
        """ Getter for image_cache_max_time """
        return self.config_vars['image_cache_max_time']

    def set_image_cache_max_time(self, value):
        """ Setter for image_cache_max_time """
        self.config_vars['image_cache_max_time'] = value

    def get_last_opened_file(self):
        """ Getter for last_opened_file """
        return self.config_vars['last_opened_file']

    def set_last_opened_file(self, value):
        """ Setter for last_opened_file """
        self.config_vars['last_opened_file'] = value

    def get_cache_folder(self):
        """ Getter for cache_folder """
        return self.config_vars['cache_folder']

    def set_cache_folder(self, value):
        """ Setter for cache_folder """
        self.config_vars['cache_folder'] = value

    def get_results_folder(self):
        """ Getter for results_folder """
        return self.config_vars['results_folder']

    def set_results_folder(self, value):
        """ Setter for results_folder """
        self.config_vars['results_folder'] = value

    # use properties for keeping transparent access to these properties
    debug_mode = property(get_debug_mode, set_debug_mode)
    never_update_data_files = property(get_never_update_data_files, set_never_update_data_files)
    reopen_last_file = property(get_reopen_last_file, set_reopen_last_file)
    use_colors_for_data = property(get_use_colors_for_data, set_use_colors_for_data)
    image_cache_max_time = property(get_image_cache_max_time, set_image_cache_max_time)
    last_opened_file = property(get_last_opened_file, set_last_opened_file)

    # following vars are not set as properties to avoid storing absolute paths in the config file
    # getter and setter could be used to store relative paths
    # cache_folder = property(get_cache_folder, set_cache_folder)
    # results_folder = property(get_results_folder, set_results_folder)

    def init_gui_config_file(self, config_file: str = '') -> None:
        """
        Initialize the config file for the gui
        :param config_file: the path to the config file to use
        """
        if config_path := os.environ.get('XDG_CONFIG_HOME'):
            self.path = os.path.join(config_path, 'UEVaultManager')
        else:
            self.path = os.path.expanduser('~/.config/UEVaultManager')
        if not os.path.isdir(self.path):
            os.makedirs(self.path)
        if config_file:
            if os.path.exists(config_file):
                self.config_path = os.path.abspath(config_file)
            else:
                self.config_path = os.path.join(self.path, clean_filename(config_file))
            log(f'UEVMGui is using non-default config file "{self.config_path}"')
        else:
            self.config_path = os.path.join(self.path, 'config_gui.ini')

        # try loading config
        try:
            self.config.read(self.config_path)
        except Exception as error:
            log('Unable to read configuration file, please ensure that file is valid! '
                f'(Error: {repr(error)})')
            log('Continuing with blank config in safe-mode...')
            self.config.read_only = True

        config_defaults = {
            'debug_mode': {
                'comment': 'Set to True to print debug information (GUI related only',
                'value': 'False'
            },
            'never_update_data_files': {
                'comment': 'Set to True to speed the update process by not updating the metadata files. FOR TESTING ONLY',
                'value': 'False'
            },
            'reopen_last_file': {
                'comment': 'Set to True to re-open the last file at startup if no input file is given',
                'value': 'True'
            },
            'use_colors_for_data': {
                'comment': 'Set to True to enable cell coloring depending on its content.It could slow down data and display refreshing',
                'value': 'True'
            },
            'last_opened_file': {
                'comment': 'File name of the last opened file',
                'value': ''
            },
            'image_cache_max_time': {
                'comment': 'Delay in seconds when image cache will be invalidated. Default value represent 15 days',
                'value': str(60 * 60 * 24 * 15)
            },
            'cache_folder': {
                'comment': 'Folder (relative or absolute) to store cached data for assets (mainly preview images)',
                'value': '../../../cache'
            },
            'results_folder': {
                'comment': 'Folder (relative or absolute) to store result files to read and save data from',
                'value': '../../../results'
            },
        }

        has_changed = False
        if 'UEVaultManager' not in self.config:
            self.config.add_section('UEVaultManager')
            has_changed = True

        for option, content in config_defaults.items():
            if not self.config.has_option('UEVaultManager', option):
                self.config.set('UEVaultManager', f';{content["comment"]}')
                self.config.set('UEVaultManager', option, content['value'])
                has_changed = True

        if has_changed:
            self.save_config_file(save_config_var=False)

    def store_config_properties(self) -> None:
        """
        store the properties in the config file
        """
        for key, value in self.config_vars.items():
            self.config.set('UEVaultManager', key, str(value))

    def save_config_file(self, save_config_var=True) -> None:
        """
        Save the config file
        """
        if save_config_var:
            self.store_config_properties()
            self.config.modified = True

        # do not save if in read-only mode or file hasn't changed
        if self.config.read_only or not self.config.modified:
            return

        # if config file has been modified externally, back-up the user-modified version before writing
        if os.path.exists(self.config_path):
            if (mod_time := int(os.stat(self.config_path).st_mtime)) != self.config.mod_time:
                new_filename = f'config.{mod_time}.ini'
                log(
                    f'Configuration file has been modified while UEVaultManager was running\nUser-modified config will be renamed to "{new_filename}"...'
                )
                os.rename(self.config_path, os.path.join(os.path.dirname(self.config_path), new_filename))

        with open(self.config_path, 'w') as cf:
            self.config.write(cf)



# coding: utf-8

import logging
from pathlib import Path

logger = logging.getLogger('LFS Utils')


def clean_filename(filename: str) -> str:
    """
    Clean a filename from invalid characters
    :param filename: The filename to clean
    :return:  The cleaned filename
    """
    return ''.join(i for i in filename if i not in '<>:"/\\|?*')


def get_dir_size(path: str) -> int:
    """
    Get the size of a directory
    :param path: The path to the directory
    :return:
    """
    return sum(f.stat().st_size for f in Path(path).glob('**/*') if f.is_file())



# coding=utf-8
"""
Implementation for:
- UEVMGuiHiddenRoot a hidden root window for the application
"""
import ttkbootstrap as ttk


class UEVMGuiHiddenRoot(ttk.Window):
    """
    This class is used to create a hidden root window for the application.
    Usefull for creating a window that is not visible to the user, but still has the ability to child windows
    """

    def __init__(self):
        super().__init__()
        self.title('UEVMGui Hidden window')
        self.geometry('100x150')
        self.withdraw()



# coding=utf-8
"""
Implementation for:
- UEVMGui the main window of the application
"""
import os
import tkinter as tk
from tkinter import filedialog as fd

import ttkbootstrap as ttk
from ttkbootstrap.constants import *

import UEVaultManager.tkgui.modules.functions as gui_f  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.functions_no_deps as gui_fn  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
from UEVaultManager.tkgui.modules.EditableTableClass import EditableTable
from UEVaultManager.tkgui.modules.functions_no_deps import set_custom_style
from UEVaultManager.tkgui.modules.TaggedLabelFrameClass import TaggedLabelFrame, WidgetType


class UEVMGui(tk.Tk):
    """
    This class is used to create the main window for the application.
    :param title: The title
    :param width: The width
    :param height: The height
    :param icon: The icon
    :param screen_index: The screen index
    :param file: The file where the data is stored or read from
    :param show_open_file_dialog: If True, the open file dialog will be shown at startup

    """

    def __init__(self, title: str, width=1200, height=800, icon='', screen_index=0, file=None, show_open_file_dialog=False, rebuild_data=False,):
        super().__init__()

        self.title(title)
        style = set_custom_style(gui_g.s.theme_name, gui_g.s.theme_font)
        self.style = style
        geometry = gui_fn.center_window_on_screen(screen_index, height, width)
        self.geometry(geometry)
        gui_fn.set_icon_and_minmax(self, icon)
        self.resizable(True, False)
        self.editable_table = None
        self.do_not_launch_search = False
        pack_def_options = {'ipadx': 5, 'ipady': 5, 'padx': 3, 'pady': 3}

        table_frame = self.TableFrame(self)

        # gui_g.UEVM_gui_ref = self  # important ! Must be donne before any use of a ProgressWindow. If not, an UEVMGuiHiddenRootClass will be created and the ProgressWindow still be displayed after the init
        self.editable_table = EditableTable(container_frame=table_frame, file=file, rows_per_page=36, show_statusbar=True)
        self.editable_table.set_preferences(gui_g.s.datatable_default_pref)

        # done in the rebuild_data() method
        self.editable_table.show()
        self.editable_table.show_page(0)

        self.table_frame = table_frame
        toolbar_frame = self.ToolbarFrame(self)
        self.toolbar_frame = toolbar_frame
        control_frame = self.ControlFrame(self)
        self.control_frame = control_frame
        options_frame = self.OptionsFrame(self)
        self.options_frame = options_frame
        self.editable_table.set_colors()

        toolbar_frame.pack(**pack_def_options, fill=tk.X, side=tk.TOP, anchor=tk.NW)
        table_frame.pack(**pack_def_options, fill=tk.BOTH, side=tk.LEFT, anchor=tk.NW, expand=True)
        control_frame.pack(**pack_def_options, fill=tk.BOTH, side=tk.RIGHT, anchor=tk.NW)
        # not displayed at start
        # options_frame.pack(**pack_def_options, fill=tk.BOTH, side=tk.RIGHT, anchor=tk.NW)

        self.bind('<Key>', self.on_key_press)
        # Bind the table to the mouse motion event
        self.editable_table.bind('<Motion>', self.on_mouse_over_cell)
        self.editable_table.bind('<Leave>', self.on_mouse_leave_cell)
        self.editable_table.bind('<<CellSelectionChanged>>', self.on_selection_change)
        self.protocol('WM_DELETE_WINDOW', self.on_close)

        if show_open_file_dialog:
            if self.load_file() == '':
                gui_f.log_error('This application could not run without a file to read data from')
                self.quit()

        # Quick edit the first row
        self.editable_table.update_quick_edit(quick_edit_frame=self.control_frame.lbtf_quick_edit, row=0)

        if rebuild_data or self.editable_table.must_rebuild:
            if gui_f.box_okcancel('Data are invalid or empty. They will be rebuilt from sources files. Do you want to continue ?'):
                if not self.editable_table.rebuild_data():
                    gui_f.log_error('Rebuild data error. This application could not run without a file to read from or some data to build from it')
                    self.destroy()  # self.quit() won't work here
                    return
            else:
                self.destroy()  # self.quit() won't work here

    class ToolbarFrame(ttk.Frame):
        """
        This class is used to create the toolbar frame
        :param container: The parent container
        """

        def __init__(self, container):
            super().__init__()
            pack_def_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.BOTH, 'expand': False}
            lblf_def_options = {'ipadx': 1, 'ipady': 1, 'expand': False}

            lblf_navigation = ttk.LabelFrame(self, text='Navigation')
            lblf_navigation.pack(side=tk.LEFT, **lblf_def_options)
            # (bootstyle is not recognized by PyCharm)
            # noinspection PyArgumentList
            btn_toggle_pagination = ttk.Button(lblf_navigation, text='Disable Pagination', command=container.toggle_pagination)
            # noinspection PyArgumentList
            btn_toggle_pagination.pack(**pack_def_options, side=tk.LEFT)
            btn_first_page = ttk.Button(lblf_navigation, text='First Page', command=container.show_first_page)
            btn_first_page.pack(**pack_def_options, side=tk.LEFT)
            btn_first_page.config(state=tk.DISABLED)
            btn_prev_page = ttk.Button(lblf_navigation, text='Prev Page', command=container.show_prev_page)
            btn_prev_page.pack(**pack_def_options, side=tk.LEFT)
            btn_prev_page.config(state=tk.DISABLED)
            entry_page_num_var = tk.StringVar(value=container.editable_table.current_page + 1)
            entry_page_num = ttk.Entry(lblf_navigation, width=5, justify=tk.CENTER, textvariable=entry_page_num_var)
            entry_page_num.pack(**pack_def_options, side=tk.LEFT)
            lbl_page_count = ttk.Label(lblf_navigation, text=f' / {container.editable_table.total_pages}')
            lbl_page_count.pack(**pack_def_options, side=tk.LEFT)
            btn_next_page = ttk.Button(lblf_navigation, text='Next Page', command=container.show_next_page)
            btn_next_page.pack(**pack_def_options, side=tk.LEFT)
            btn_last_page = ttk.Button(lblf_navigation, text='Last Page', command=container.show_last_page)
            btn_last_page.pack(**pack_def_options, side=tk.LEFT)
            btn_prev = ttk.Button(lblf_navigation, text='Prev Asset', command=container.prev_asset)
            btn_prev.pack(**pack_def_options, side=tk.LEFT)
            btn_next = ttk.Button(lblf_navigation, text='Next Asset', command=container.next_asset)
            btn_next.pack(**pack_def_options, side=tk.RIGHT)

            lblf_display = ttk.LabelFrame(self, text='Display')
            lblf_display.pack(side=tk.LEFT, **lblf_def_options)
            btn_expand = ttk.Button(lblf_display, text='Expand Cols', command=container.editable_table.expand_columns)
            btn_expand.pack(**pack_def_options, side=tk.LEFT)
            btn_shrink = ttk.Button(lblf_display, text='Shrink Cols', command=container.editable_table.contract_columns)
            btn_shrink.pack(**pack_def_options, side=tk.LEFT)
            btn_autofit = ttk.Button(lblf_display, text='Autofit Cols', command=container.editable_table.autofit_columns)
            btn_autofit.pack(**pack_def_options, side=tk.LEFT)
            btn_zoom_in = ttk.Button(lblf_display, text='Zoom In', command=container.editable_table.zoom_in)
            btn_zoom_in.pack(**pack_def_options, side=tk.LEFT)
            btn_zoom_out = ttk.Button(lblf_display, text='Zoom Out', command=container.editable_table.zoom_out)
            btn_zoom_out.pack(**pack_def_options, side=tk.LEFT)

            lblf_commands = ttk.LabelFrame(self, text='Cli commands')
            lblf_commands.pack(side=tk.LEFT, **lblf_def_options)
            btn_help = ttk.Button(lblf_commands, text='Help', command=lambda: container.run_cli_command('print_help'))
            btn_help.pack(**pack_def_options, side=tk.LEFT)
            btn_status = ttk.Button(lblf_commands, text='Status', command=lambda: container.run_cli_command('status'))
            btn_status.pack(**pack_def_options, side=tk.LEFT)
            btn_info = ttk.Button(lblf_commands, text='Info', command=lambda: container.run_cli_command('info'))
            btn_info.pack(**pack_def_options, side=tk.LEFT)
            btn_info = ttk.Button(lblf_commands, text='List Files', command=lambda: container.run_cli_command('list_files'))
            btn_info.pack(**pack_def_options, side=tk.LEFT)

            lblf_actions = ttk.LabelFrame(self, text='Actions')
            lblf_actions.pack(side=tk.RIGHT, **lblf_def_options)
            # noinspection PyArgumentList
            btn_toggle_options = ttk.Button(lblf_actions, text='Show Options', command=container.toggle_options_pane, state=tk.DISABLED)
            btn_toggle_options.pack(**pack_def_options, side=tk.LEFT)
            # noinspection PyArgumentList
            btn_toggle_controls = ttk.Button(lblf_actions, text='Hide Controls', command=container.toggle_controls_pane)
            btn_toggle_controls.pack(**pack_def_options, side=tk.LEFT)
            # noinspection PyArgumentList
            btn_on_close = ttk.Button(lblf_actions, text='Quit', command=container.on_close, bootstyle=WARNING)
            btn_on_close.pack(**pack_def_options, side=tk.RIGHT)

            # Bind events for the Entry widget
            entry_page_num.bind('<FocusOut>', container.on_entry_page_num_changed)
            entry_page_num.bind('<Return>', container.on_entry_page_num_changed)

            self.btn_toggle_pagination = btn_toggle_pagination
            self.btn_first_page = btn_first_page
            self.btn_prev_page = btn_prev_page
            self.btn_next_page = btn_next_page
            self.btn_last_page = btn_last_page
            self.btn_toggle_options = btn_toggle_options
            self.btn_toggle_controls = btn_toggle_controls
            self.lbl_page_count = lbl_page_count
            self.entry_page_num = entry_page_num
            self.entry_page_num_var = entry_page_num_var

    class TableFrame(ttk.Frame):
        """
        The TableFrame is a container for the Table widget.
        :param container: The parent container.
        """

        def __init__(self, container):
            super().__init__(container)
            self.container = container

    class ControlFrame(ttk.Frame):
        """
        The ControlFrame is a container for the filter controls.
        :param container: The parent container.
        """

        def __init__(self, container):
            super().__init__()

            # grid_def_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'sticky': tk.NW}
            pack_def_options = {'ipadx': 2, 'ipady': 2, 'fill': tk.BOTH, 'expand': False}
            grid_fw_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'sticky': tk.EW}  # full width
            lblf_def_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.X, 'expand': False}
            lblf_fw_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.X, 'expand': True}  # full width

            cat_vars = container.update_category_var()
            grab_results = cat_vars['grab_results']
            categories = cat_vars['categories']
            lblf_content = ttk.LabelFrame(self, text='Content')
            lblf_content.pack(**lblf_def_options)
            btn_edit_row = ttk.Button(lblf_content, text='Edit Row', command=container.editable_table.create_edit_record_window)
            btn_edit_row.grid(row=0, column=0, **grid_fw_options)
            btn_reload_data = ttk.Button(lblf_content, text='Reload File Content', command=container.reload_data)
            btn_reload_data.grid(row=0, column=1, **grid_fw_options)
            btn_rebuild_file = ttk.Button(lblf_content, text='Rebuild File Content', command=container.rebuild_data)
            btn_rebuild_file.grid(row=0, column=2, **grid_fw_options)
            lblf_content.columnconfigure('all', weight=1)  # important to make the buttons expand

            lbf_filter_cat = ttk.LabelFrame(self, text='Search and Filter')
            # row 0
            cur_col = 0
            cur_row = 0
            lbf_filter_cat.pack(fill=tk.X, anchor=tk.NW, ipadx=5, ipady=5)
            var_is_purchased = tk.BooleanVar(value=False)
            var_is_purchased.trace_add('write', container.on_check_change)
            ck_purchased = ttk.Checkbutton(lbf_filter_cat, text='Purchased', variable=var_is_purchased)
            ck_purchased.grid(row=cur_row, column=cur_col, **grid_fw_options)
            cur_col += 1
            var_must_buy = tk.BooleanVar(value=False)
            var_must_buy.trace_add('write', container.on_check_change)
            ck_must_buy = ttk.Checkbutton(lbf_filter_cat, text='Must buy', variable=var_must_buy)
            ck_must_buy.grid(row=cur_row, column=cur_col, **grid_fw_options)
            cur_col += 1
            var_on_sale = tk.BooleanVar(value=False)
            var_on_sale.trace_add('write', container.on_check_change)
            ck_on_sale = ttk.Checkbutton(lbf_filter_cat, text='On sale', variable=var_on_sale)
            ck_on_sale.grid(row=cur_row, column=cur_col, **grid_fw_options)
            cur_col += 1
            var_is_not_obsolete = tk.BooleanVar(value=False)
            var_is_not_obsolete.trace_add('write', container.on_check_change)
            ck_obsolete = ttk.Checkbutton(lbf_filter_cat, text='Not obsolete', variable=var_is_not_obsolete)
            ck_obsolete.grid(row=cur_row, column=cur_col, **grid_fw_options)
            # row 1
            cur_row += 1
            cur_col = 0
            lbl_category = ttk.Label(lbf_filter_cat, text='Category')
            lbl_category.grid(row=cur_row, column=cur_col, **grid_fw_options)
            cur_col += 1
            var_category = tk.StringVar(value=categories[0])
            opt_category = ttk.Combobox(lbf_filter_cat, textvariable=var_category, values=categories)
            opt_category.grid(row=cur_row, column=cur_col, columnspan=3, **grid_fw_options)
            # row 2
            cur_row += 1
            cur_col = 0
            lbl_grab_results = ttk.Label(lbf_filter_cat, text='Grab result')
            lbl_grab_results.grid(row=cur_row, column=cur_col, **grid_fw_options)
            cur_col += 1
            var_grab_results = tk.StringVar(value=grab_results[0])
            opt_grab_results = ttk.Combobox(lbf_filter_cat, textvariable=var_grab_results, values=grab_results)
            opt_grab_results.grid(row=cur_row, column=cur_col, columnspan=3, **grid_fw_options)
            # row 3
            cur_row += 1
            cur_col = 0
            var_global_search = tk.StringVar(value=gui_g.s.default_global_search)
            entry_search = ttk.Entry(lbf_filter_cat, textvariable=var_global_search)
            entry_search.grid(row=cur_row, column=cur_col, columnspan=2, **grid_fw_options)
            entry_search.bind('<FocusIn>', self.del_entry_search)
            # entry_search.bind('<FocusOut>', container.search)
            cur_col += 2
            btn_filter_by_text = ttk.Button(lbf_filter_cat, text='Apply All', command=container.apply_filters)
            btn_filter_by_text.grid(row=cur_row, column=cur_col, **grid_fw_options)
            cur_col += 1
            btn_reset_search = ttk.Button(lbf_filter_cat, text='Reset All', command=container.reset_filters)
            btn_reset_search.grid(row=cur_row, column=cur_col, **grid_fw_options)
            lbf_filter_cat.columnconfigure('all', weight=1)  # important to make the buttons expand

            lblf_files = ttk.LabelFrame(self, text='Files')
            lblf_files.pack(**lblf_def_options)
            lbl_file_name = ttk.Label(lblf_files, text='Current File: ')
            lbl_file_name.grid(row=0, column=0, columnspan=3, **grid_fw_options)
            var_entry_file_name = tk.StringVar(value=container.editable_table.file)
            entry_file_name = ttk.Entry(lblf_files, textvariable=var_entry_file_name, state='readonly')
            entry_file_name.grid(row=1, column=0, columnspan=3, **grid_fw_options)
            btn_save_file = ttk.Button(lblf_files, text='Save to File', command=container.save_file)
            btn_save_file.grid(row=2, column=0, **grid_fw_options)
            btn_export_button = ttk.Button(lblf_files, text='Export Selection', command=container.export_selection)
            btn_export_button.grid(row=2, column=1, **grid_fw_options)
            btn_load_file = ttk.Button(lblf_files, text='Load a file', command=container.load_file)
            btn_load_file.grid(row=2, column=2, **grid_fw_options)
            lblf_files.columnconfigure('all', weight=1)  # important to make the buttons expand

            # note : the TAG of the child widgets of the lbf_quick_edit will also be used in the editable_table.quick_edit method
            # to get the widgets it needs. So they can't be changed freely
            lbtf_quick_edit = TaggedLabelFrame(self, text='Quick Edit User fields')
            lbtf_quick_edit.pack(**lblf_fw_options, anchor=tk.NW)
            lbl_desc = ttk.Label(lbtf_quick_edit, text='Changing this values will change the values of \nthe selected row when losing focus')
            lbl_desc.pack(**pack_def_options)
            lbtf_quick_edit.add_child(widget_type=WidgetType.ENTRY, tag='Url', focus_out_callback=container.on_quick_edit_focus_out)
            lbtf_quick_edit.add_child(
                widget_type=WidgetType.TEXT, tag='Comment', focus_out_callback=container.on_quick_edit_focus_out, width=10, height=4
            )
            lbtf_quick_edit.add_child(widget_type=WidgetType.ENTRY, tag='Stars', focus_out_callback=container.on_quick_edit_focus_out)
            lbtf_quick_edit.add_child(widget_type=WidgetType.ENTRY, tag='Test result', focus_out_callback=container.on_quick_edit_focus_out)
            lbtf_quick_edit.add_child(
                widget_type=WidgetType.CHECKBUTTON,
                tag='Must buy',
                label='',
                images_folder=gui_g.s.assets_folder,
                click_on_callback=container.on_switch_edit_flag,
                default_content=False
            )
            lbtf_quick_edit.add_child(widget_type=WidgetType.ENTRY, tag='Installed folder', default_content='Installed in')
            lbtf_quick_edit.add_child(widget_type=WidgetType.ENTRY, tag='Origin', focus_out_callback=container.on_quick_edit_focus_out)
            lbtf_quick_edit.add_child(widget_type=WidgetType.ENTRY, tag='Alternative', focus_out_callback=container.on_quick_edit_focus_out)

            lbt_image_preview = ttk.LabelFrame(self, text='Image Preview')
            lbt_image_preview.pack(**lblf_fw_options, anchor=tk.SW)
            canvas_image = tk.Canvas(lbt_image_preview, width=gui_g.s.preview_max_width, height=gui_g.s.preview_max_height, highlightthickness=0)
            canvas_image.pack(side=tk.BOTTOM, expand=True, anchor=tk.CENTER)
            canvas_image.create_rectangle((0, 0), (gui_g.s.preview_max_width, gui_g.s.preview_max_height), fill='black')

            lblf_bottom = ttk.Frame(self)
            lblf_bottom.pack(**lblf_def_options)
            ttk.Sizegrip(lblf_bottom).pack(side=tk.RIGHT)

            # store the controls that need to be accessible outside the class
            self.var_entry_file_name = var_entry_file_name
            self.var_category = var_category
            self.var_global_search = var_global_search
            self.var_is_purchased = var_is_purchased
            self.var_is_not_obsolete = var_is_not_obsolete
            self.var_must_buy = var_must_buy
            self.var_on_sale = var_on_sale
            self.var_grab_results = var_grab_results

            self.entry_search = entry_search
            self.lbtf_quick_edit = lbtf_quick_edit
            self.canvas_image = canvas_image

        def reset_entry_search(self, _event=None) -> None:
            """
            Reset the search entry to the default text.
            :param _event:
            """
            self.entry_search.delete(0, 'end')
            self.entry_search.insert(0, gui_g.s.default_global_search)

        def del_entry_search(self, _event=None) -> None:
            """
            Delete the text in the search entry.
            :param _event:
            """
            self.entry_search.delete(0, 'end')

    class OptionsFrame(ttk.Frame):
        """
        This class is used to create the toolbar frame
        :param _container: The parent container
        """

        def __init__(self, _container):
            super().__init__()
            # pack_def_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'fill': tk.BOTH, 'expand': False}
            lblf_def_options = {'ipadx': 1, 'ipady': 1, 'expand': False}
            grid_fw_options = {'ipadx': 2, 'ipady': 2, 'padx': 2, 'pady': 2, 'sticky': tk.EW}  # full width

            lblf_options = ttk.LabelFrame(self, text='Command Options')
            lblf_options.pack(side=tk.TOP, **lblf_def_options)
            # row 0
            cur_col = 0
            cur_row = 0
            force_refresh_var = tk.BooleanVar(value=gui_g.UEVM_cli_args.get('force', False))
            force_refresh_var.trace_add('write', lambda name, index, mode: gui_g.set_args_force_refresh(force_refresh_var.get()))
            ck_force_refresh = ttk.Checkbutton(lblf_options, text='Force refresh', variable=force_refresh_var)
            ck_force_refresh.grid(row=cur_row, column=cur_col, **grid_fw_options)
            # row 1
            cur_row += 1
            cur_col = 0
            offline_var = tk.BooleanVar(value=gui_g.UEVM_cli_args.get('offli', False))
            offline_var.trace_add('write', lambda name, index, mode: gui_g.set_args_offline(offline_var.get()))
            ck_offline = ttk.Checkbutton(lblf_options, text='Offline Mode', variable=offline_var)
            ck_offline.grid(row=cur_row, column=cur_col, **grid_fw_options)
            # row 2
            cur_row += 1
            cur_col = 0
            debug_var = tk.BooleanVar(value=gui_g.UEVM_cli_args.get('debug', False))
            debug_var.trace_add('write', lambda name, index, mode: gui_g.set_args_debug(debug_var.get()))
            ck_debug = ttk.Checkbutton(lblf_options, text='Debug mode', variable=debug_var)
            ck_debug.grid(row=cur_row, column=cur_col, **grid_fw_options)

    def _open_file_dialog(self, save_mode=False, filename=None) -> str:
        """
        Open a file dialog to choose a file to save or load data to/from
        :param save_mode: if True, the dialog will be in saving mode, else in loading mode
        :param filename: the default filename to use
        :return: the chosen filename
        """
        # adding category to the default filename
        if not filename:
            filename = gui_g.s.default_filename
        initial_dir = os.path.dirname(filename)
        default_filename = os.path.basename(filename)  # remove dir
        default_ext = os.path.splitext(default_filename)[1]  # get extension
        default_filename = os.path.splitext(default_filename)[0]  # get filename without extension
        try:
            # if the file is empty or absent or invalid when creating the class, the control_frame is not defined
            category = self.control_frame.var_category.get().replace('/', '_')
        except AttributeError:
            category = None
        if category and category != gui_g.s.default_category_for_all:
            default_filename = default_filename + '_' + category + default_ext
        else:
            default_filename = default_filename + default_ext
        if save_mode:
            filename = fd.asksaveasfilename(
                title='Choose a file to save data to', initialdir=initial_dir, filetypes=gui_g.s.data_filetypes, initialfile=default_filename
            )
        else:
            filename = fd.askopenfilename(
                title='Choose a file to read data from', initialdir=initial_dir, filetypes=gui_g.s.data_filetypes, initialfile=default_filename
            )
        return filename

    def _change_navigation_state(self, state: str) -> None:
        """
        Change the state of the navigation buttons
        :param state: 'normal' or 'disabled'
        """
        self.toolbar_frame.btn_first_page.config(state=state)
        self.toolbar_frame.btn_prev_page.config(state=state)
        self.toolbar_frame.btn_next_page.config(state=state)
        self.toolbar_frame.btn_last_page.config(state=state)
        self.toolbar_frame.entry_page_num.config(state=state)

    def on_key_press(self, event) -> None:
        """
        Handle key press events
        :param event:
        """
        if event.keysym == 'Escape':
            if gui_g.edit_cell_window_ref:
                gui_g.edit_cell_window_ref.quit()
                gui_g.edit_cell_window_ref = None
            elif gui_g.edit_row_window_ref:
                gui_g.edit_row_window_ref.quit()
                gui_g.edit_row_window_ref = None
            else:
                self.on_close()
        # elif event.keysym == 'Return':
        #    self.editable_table.create_edit_record_window()

    def on_mouse_over_cell(self, event=None) -> None:
        """
        Show the image of the asset when the mouse is over the cell
        :param event:
        """
        if event is None:
            return
        canvas_image = self.control_frame.canvas_image
        try:
            image_url = self.editable_table.get_image_url(row=self.editable_table.get_row_clicked(event))
            gui_f.show_asset_image(image_url=image_url, canvas_image=canvas_image)
        except IndexError:
            gui_f.show_default_image(canvas_image)

    def on_mouse_leave_cell(self, _event=None) -> None:
        """
        Show the default image when the mouse leaves the cell
        :param _event:
        """
        canvas_image = self.control_frame.canvas_image
        gui_f.show_default_image(canvas_image=canvas_image)

    def on_selection_change(self, event=None) -> None:
        """
        When the selection changes, show the selected row in the quick edit frame
        :param event:
        """
        selected_row = event.widget.currentrow
        self.editable_table.update_quick_edit(quick_edit_frame=self.control_frame.lbtf_quick_edit, row=selected_row)

    def on_entry_page_num_changed(self, _event=None) -> None:
        """
        When the page number changes, show the corresponding page
        :param _event:
        """
        page_num = 0
        try:
            page_num = self.toolbar_frame.entry_page_num.get()
            page_num = int(page_num)
            page_num -= 1
            gui_f.log_debug(f'showing page {page_num}')
            self.editable_table.show_page(page_num)
        except (ValueError, UnboundLocalError) as error:
            gui_f.log_error(f'could not convert page number {page_num} to int. {error!r}')

    # noinspection PyUnusedLocal
    def on_quick_edit_focus_out(self, event=None, tag='') -> None:
        """
        When the focus leaves a quick edit widget, save the value
        :param event:
        :param tag: tag of the widget that triggered the event
        """
        if tag == '':
            return
        widget = self.control_frame.lbtf_quick_edit.get_child_by_tag(tag)
        if widget is None:
            gui_f.log_warning(f'Could not find a widget with tag {tag}')
            return
        col = widget.col
        row = widget.row
        if col is None or row is None or col < 0 or row < 0:
            gui_f.log_debug(f'invalid values for row={row} and col={col}')
            return
        value = widget.get_content()
        self.editable_table.quick_edit_save_value(col=col, row=row, value=value)

    # noinspection PyUnusedLocal
    def on_switch_edit_flag(self, event=None, tag='') -> None:
        """
        When the focus leaves a quick edit widget, save the value
        :param event: event that triggered the call
        :param tag: tag of the widget that triggered the event
        """
        if tag == '':
            return
        widget = self.control_frame.lbtf_quick_edit.get_child_by_tag(tag)
        if widget is None or widget.widget_type != WidgetType.CHECKBUTTON:
            gui_f.log_warning(f'Could not find a CHECKBUTTON widget with tag {tag}')
            return
        col = widget.col
        row = widget.row
        if col is None or row is None or col < 0 or row < 0:
            gui_f.log_debug(f'invalid values for row={row} and col={col}')
            return

        value = widget.switch_state(event=event)
        self.editable_table.quick_edit_save_value(col=col, row=row, value=value)

    # noinspection PyUnusedLocal
    def on_check_change(self, *args) -> None:
        """
        When a checkbutton is changed, launch a search
        :param args:
        """
        if not self.do_not_launch_search:
            self.apply_filters()

    def on_close(self) -> None:
        """
        When the window is closed, check if there are unsaved changes and ask the user if he wants to save them
        """
        if self.editable_table is not None and self.editable_table.must_save:
            if gui_f.box_yesno('Changes have been made. Do you want to save them in the source file ?'):
                self.save_file(show_dialog=False)
        if gui_g.s.reopen_last_file:
            gui_g.s.last_opened_file = self.editable_table.file
        gui_g.s.save_config_file()
        self.quit()

    def load_file(self) -> str:
        """
        Load a file
        :return: the name of the file that was loaded
        """
        filename = self._open_file_dialog(filename=self.editable_table.file)
        if filename and os.path.isfile(filename):
            self.editable_table.file = filename
            self.editable_table.load_data()
            self.editable_table.show_page(0)
            self.update_page_numbers()
            self.update_file_name()
            gui_f.box_message(f'The file {filename} as been read')
        return filename

    def save_file(self, show_dialog=True) -> str:
        """
        Save the data to a file
        :param show_dialog: if True, show a dialog to select the file to save to, if False, use the current file
        """
        if show_dialog:
            filename = self._open_file_dialog(filename=self.editable_table.file, save_mode=True)
            if filename:
                self.editable_table.file = filename
        else:
            filename = self.editable_table.file

        if filename:
            self.editable_table.save_data()
            self.update_file_name()
            gui_f.box_message(f'Data Saved to {self.editable_table.file}')
        return filename

    def export_selection(self) -> None:
        """
        Export the selected rows to a file
        """
        # Get selected row indices
        selected_row_indices = self.editable_table.multiplerowlist
        if selected_row_indices:
            selected_rows = self.editable_table.data_filtered.iloc[selected_row_indices]
            filename = self._open_file_dialog(save_mode=True, filename=self.editable_table.file)
            if filename:
                selected_rows.to_csv(filename, index=False)
                gui_f.box_message(f'Selected rows exported to "{filename}"')
        else:
            gui_f.box_message('Select at least one row first')

    def apply_filters(self, _event=None) -> None:
        """
        Search for a string in the table
        :param _event: Event
        """
        search_text = self.control_frame.var_global_search.get()
        category = self.control_frame.var_category.get()
        grab_results = self.control_frame.var_grab_results.get()
        purchased = self.control_frame.var_is_purchased.get()
        not_obsolete = self.control_frame.var_is_not_obsolete.get()
        must_buy = self.control_frame.var_must_buy.get()
        on_sale = self.control_frame.var_on_sale.get()
        gui_g.UEVM_filter_category = category if category != gui_g.s.default_category_for_all else ''
        self.toggle_pagination(forced_value=False)
        filter_dict = {}
        if category != gui_g.s.default_category_for_all:
            filter_dict['Category'] = category
        else:
            filter_dict.pop('Category', None)
        if grab_results != gui_g.s.default_category_for_all:
            filter_dict['Grab result'] = grab_results
        else:
            filter_dict.pop('Grab result', None)
        if purchased:
            filter_dict['Purchased'] = True
        else:
            filter_dict.pop('Purchased', None)
        if not_obsolete:
            filter_dict['Obsolete'] = False
            filter_dict['Status'] = 'active'
        else:
            filter_dict.pop('Obsolete', None)
            filter_dict.pop('Status', None)
        if must_buy:
            filter_dict['Must buy'] = True
        else:
            filter_dict.pop('Must buy', None)
        if on_sale:
            filter_dict['On sale'] = True
        else:
            filter_dict.pop('On sale', None)
        self.editable_table.apply_filters(filter_dict, global_search=search_text)
        # self.control_frame.reset_entry_search()

    def reset_filters(self) -> None:
        """
        Reset the search controls to their default values
        """
        self.control_frame.var_global_search.set(gui_g.s.default_global_search)
        self.control_frame.var_category.set(gui_g.s.default_category_for_all)
        gui_g.UEVM_filter_category = ''
        self.do_not_launch_search = True  # Prevent the search to be launched when the checkbuttons are changed
        self.control_frame.var_is_purchased.set(False)
        self.control_frame.var_is_not_obsolete.set(False)
        self.do_not_launch_search = False
        self.editable_table.reset_filters()

    def toggle_pagination(self, forced_value=None) -> None:
        """
        Toggle pagination. Will change the navigation buttons states when pagination is changed
        :param forced_value: if not None, will force the pagination to the given value
        """
        if forced_value is not None:
            self.editable_table.pagination_enabled = forced_value
        else:
            self.editable_table.pagination_enabled = not self.editable_table.pagination_enabled
        self.editable_table.show_page()
        if not self.editable_table.pagination_enabled:
            # Disable prev/next buttons when pagination is disabled
            self._change_navigation_state(tk.DISABLED)
            self.toolbar_frame.btn_toggle_pagination.config(text='Enable  Pagination')
        else:
            self.update_page_numbers()  # will also update buttons status
            self.toolbar_frame.btn_toggle_pagination.config(text='Disable Pagination')

    def show_first_page(self) -> None:
        """
        Show the first page of the table
        """
        self.editable_table.first_page()
        self.update_page_numbers()

    def show_prev_page(self) -> None:
        """
        Show the previous page of the table
        """
        self.editable_table.prev_page()
        self.update_page_numbers()

    def show_next_page(self) -> None:
        """
        Show the next page of the table
        """
        self.editable_table.next_page()
        self.update_page_numbers()

    def show_last_page(self) -> None:
        """
        Show the last page of the table
        """
        self.editable_table.last_page()
        self.update_page_numbers()

    def prev_asset(self) -> None:
        """
        Move to the previous asset in the table
        """
        self.editable_table.move_to_prev_record()

    def next_asset(self) -> None:
        """
        Move to the next asset in the table
        """
        self.editable_table.move_to_next_record()

    # noinspection DuplicatedCode
    def toggle_controls_pane(self) -> None:
        """
        Toggle the visibility of the controls pane
        """
        # noinspection DuplicatedCode
        if self.control_frame.winfo_ismapped():
            self.control_frame.pack_forget()
            self.toolbar_frame.btn_toggle_controls.config(text='Show Control')
            self.toolbar_frame.btn_toggle_options.config(state=tk.NORMAL)

        else:
            self.control_frame.pack(side=tk.RIGHT, fill=tk.BOTH)
            self.toolbar_frame.btn_toggle_controls.config(text='Hide Control')
            self.toolbar_frame.btn_toggle_options.config(state=tk.DISABLED)

    # noinspection DuplicatedCode
    def toggle_options_pane(self) -> None:
        """
        Toggle the visibility of the Options pane
        """
        # noinspection DuplicatedCode
        if self.options_frame.winfo_ismapped():
            self.options_frame.pack_forget()
            self.toolbar_frame.btn_toggle_options.config(text='Show Options')
            self.toolbar_frame.btn_toggle_controls.config(state=tk.NORMAL)
        else:
            self.options_frame.pack(side=tk.RIGHT, fill=tk.BOTH)
            self.toolbar_frame.btn_toggle_options.config(text='Hide Options')
            self.toolbar_frame.btn_toggle_controls.config(state=tk.DISABLED)

    def update_page_numbers(self) -> None:
        """
        Update the page numbers in the toolbar
        """
        page_num = self.editable_table.current_page + 1
        self.toolbar_frame.entry_page_num_var.set(page_num)
        self.toolbar_frame.lbl_page_count.config(text=f' / {self.editable_table.total_pages}')
        # enable all buttons by default
        self._change_navigation_state(tk.NORMAL)

        if not self.editable_table.pagination_enabled:
            self.toolbar_frame.entry_page_num.config(state=tk.NORMAL)
        if page_num == 1:
            self.toolbar_frame.btn_first_page.config(state=tk.DISABLED)
            self.toolbar_frame.btn_prev_page.config(state=tk.DISABLED)
        elif page_num == self.editable_table.total_pages:
            self.toolbar_frame.btn_next_page.config(state=tk.DISABLED)
            self.toolbar_frame.btn_last_page.config(state=tk.DISABLED)

    def update_file_name(self) -> None:
        """
        Update the file name in the control frame
        """
        filename = self.editable_table.file
        self.control_frame.var_entry_file_name.set(filename)

    def update_category_var(self) -> dict:
        """
        Update the category variable with the current categories in the data
        :return: a dict with the new categories list as value and the key is the name of the variable.
        """
        try:
            # if the file is empty or absent or invalid when creating the class, the data is empty, so no categories
            categories = list(self.editable_table.data['Category'].cat.categories)
        except (AttributeError, TypeError):
            categories = []
        categories.insert(0, gui_g.s.default_category_for_all)
        try:
            # if the file is empty or absent or invalid when creating the class, the data is empty, so no categories
            grab_results = list(self.editable_table.data['Grab result'].cat.categories)
        except (AttributeError, TypeError):
            grab_results = []
        grab_results.insert(0, gui_g.s.default_category_for_all)
        return {'categories': categories, 'grab_results': grab_results}

    def reload_data(self) -> None:
        """
        Reload the data from the file
        """
        if not self.editable_table.must_save or (
            self.editable_table.must_save and gui_f.box_yesno('Changes have been made, they will be lost. Are you sure you want to continue ?')
        ):
            self.editable_table.reload_data()
            gui_f.box_message(f'Data Reloaded from {self.editable_table.file}')
            self.update_page_numbers()
            self.update_category_var()

    def rebuild_data(self) -> None:
        """
        Rebuild the data from the file. Will ask for confirmation before rebuilding
        """
        if gui_f.box_yesno(
            f'The process will change the content of the windows and the {self.editable_table.file} file.\nAre you sure you want to continue ?'
        ):
            if self.editable_table.rebuild_data():
                gui_f.box_message(f'Data rebuilt from {self.editable_table.file}')
                self.update_page_numbers()
                self.update_category_var()

    def run_cli_command(self, command_name='') -> None:
        """
        Execute the 'status' command and display the result in DisplayContentWindow
        :param command_name: the name of the command to execute
        """
        if command_name == '':
            return
        if gui_g.UEVM_cli_ref is None:
            gui_f.from_cli_only_message()
            return
        row = self.editable_table.getSelectedRow()
        col = self.editable_table.model.df.columns.get_loc('App name')
        app_name = self.editable_table.model.getValueAt(row, col)

        # gui_g.UEVM_cli_args['offline'] = True  # speed up some commands
        # set default options for the cli command to execute
        gui_g.UEVM_cli_args['gui'] = True
        gui_g.UEVM_cli_args['full_help'] = True
        gui_g.UEVM_cli_args['csv'] = False  # mandatory for displaying the result in the DisplayContentWindow
        gui_g.UEVM_cli_args['tcsv'] = False  # mandatory for displaying the result in the DisplayContentWindow
        gui_g.UEVM_cli_args['json'] = False  # mandatory for displaying the result in the DisplayContentWindow

        if app_name != '':
            gui_g.UEVM_cli_args['app_name'] = app_name

        if gui_g.display_content_window_ref is None or not gui_g.display_content_window_ref.winfo_viewable():
            # we display the window only if it is not already displayed
            function_to_call = getattr(gui_g.UEVM_cli_ref, command_name)
            function_to_call(gui_g.UEVM_cli_args)



# coding=utf-8
"""
GUI module for UEVaultManager
this can be run directly by executing this script
or by the --edit command option for the UEVaultManager cli application, for instance by running `cli --edit --input <my_source_file>'
"""
import os.path

import UEVaultManager.tkgui.modules.functions_no_deps as gui_fn  # using the shortest variable name for globals for convenience
import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
from UEVaultManager.tkgui.modules.UEVMGuiClass import UEVMGui


def init_gui(open_mainwindow=True) -> str:
    """
    Main function for the GUI
    :param open_mainwindow: if True, the main window will be opened (default mode).
            Set to False for running the GUI initialization only, useful if called from cli.py
    :return: the path to the csv file to use at startup. It's used when the window is opened from the cli.py script
    """
    app_icon_filename = gui_fn.path_from_relative_to_absolute(gui_g.s.app_icon_filename)
    csv_filename = gui_fn.path_from_relative_to_absolute(gui_g.s.csv_filename)
    rebuild = False
    if not os.path.isfile(csv_filename):
        csv_filename = gui_fn.create_empty_file(csv_filename)
        rebuild = True
    if open_mainwindow:
        main_window = UEVMGui(
            title=gui_g.s.app_title,
            width=gui_g.s.app_width,
            height=gui_g.s.app_height,
            icon=app_icon_filename,
            screen_index=0,
            file=csv_filename,
            rebuild_data=rebuild
        )
        main_window.mainloop()

    return csv_filename


if __name__ == '__main__':
    init_gui()



# coding: utf-8
"""
UEVaultManager setup file.
"""
import requirements
import sys
from pathlib import Path

import setuptools
from setuptools import setup

from UEVaultManager import __name__, __version__, __codename__, __license__, __author__, __author_email__, __description__, __url__

if sys.version_info < (3, 9):
    sys.exit('python 3.9 or higher is required for UEVaultManager')

current_folder = Path(__file__).parent

# Read __long_description__ from README.md
######################
__long_description__ = Path.joinpath(current_folder, 'README.md').read_text()
# add information about the version and codename in the PyPI page because codename s not available by default
__long_description__ += f'\n\n {__name__} ## version:{__version__} ## codename: {__codename__}'
# __long_description__ = ''.join(__long_description__[8:16])  # keep only description text

# Read requirements from the requirements.txt file
######################
# Note: This can cause problems with hyphenated package names.
# It is fairly common for PyPi packages to be listed with a hyphen in their name,
# but for all other references to them to have underscores.
# see: https://github.com/pypa/setuptools/issues/1080
requirements_from_file = []
with open(Path.joinpath(current_folder, 'requirements.txt')) as fd:
    for req in requirements.parse(fd):
        if req.name:
            name = req.name.replace("-", "_")
            full_line = name + "".join(["".join(list(spec)) for spec in req.specs])
            requirements_from_file.append(full_line)
        else:
            # no name meens it's a package from a source URL , e.g.:
            # a GitHub URL doesn't have a name
            # I am not sure what to do with this, actually, right now I just ignore it...
            pass

setup(
    name=__name__,
    version=__version__,
    license=__license__,
    author=__author__,
    author_email=__author_email__,
    description=__description__,
    long_description=__long_description__,
    long_description_content_type='text/markdown',
    url=__url__,
    packages=setuptools.find_packages(),
    # data_files=[('', ['UEVaultManager/assets/UEVM_200x200.png','UEVaultManager/assets/main.ico'])],
    package_data={'': ['assets/*']},
    entry_points=dict(console_scripts=['UEVaultManager = UEVaultManager.cli:main']),
    install_requires=requirements_from_file,
    extras_require=dict(webview=['pywebview>=3.4'], webview_gtk=[
        'pywebview>=3.4',  #
        'PyGObject'
    ]),
    python_requires='>=3.9',
    classifiers=[
        'License :: OSI Approved :: BSD License',  #
        'Programming Language :: Python :: 3.9',  #
        'Programming Language :: Python :: 3.10', 'Programming Language :: Python :: 3.11',  #
        'Operating System :: OS Independent',  #
        'Development Status :: 5 - Production/Stable'
        # 'Development Status :: 4 - Beta'
    ]
)

if __name__ == '__main__':
    setup()



# coding=utf-8
"""
Class definitions for:
- HiddenAliasSubparsersAction: Subclass of argparse._SubParsersAction that hides aliases from the help output.
"""
import argparse


# noinspection PyUnresolvedReferences,PyProtectedMember
class HiddenAliasSubparsersAction(argparse._SubParsersAction):
    """
    Subclass of argparse._SubParsersAction that hides aliases from the help output.
    """

    def add_parser(self, name, **kwargs):
        # set prog from the existing prefix
        if kwargs.get('prog') is None:
            kwargs['prog'] = f'{self._prog_prefix} {name}'

        aliases = kwargs.pop('aliases', ())
        hide_aliases = kwargs.pop('hide_aliases', False)

        # create a pseudo-action to hold the choice help
        if 'help' in kwargs:
            help_value = kwargs.pop('help')
            _aliases = None if hide_aliases else aliases
            choice_action = self._ChoicesPseudoAction(name, _aliases, help_value)
            self._choices_actions.append(choice_action)

        # create the parser and add it to the map
        parser = self._parser_class(**kwargs)
        self._name_parser_map[name] = parser

        # make parser available under aliases also
        for alias in aliases:
            self._name_parser_map[alias] = parser

        return parser
